/*
 * Enumeration.java
 * Part of the B'am library - a collection of general-purpose utility classes.
 * 
 * Copyright (C)2006 Stephen Battey
 * Contact: resource@lucidviews.net
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

package net.lucidviews.util;

import java.util.List;
import java.util.NoSuchElementException;
import java.util.ArrayList;
import java.util.ListIterator;
import java.util.Iterator;


/**
 * An enumerated list of values.
 * <br />
 * An enumeration is a finite set of distinct values.
 * <br />
 * Each value in an <code>Enumeration</code> is an instance of an
 * <code>EnumValue</code>. Creating an enumeration simply requires you to
 * create a sub-class of <code>EnumValue</code> - see the documentation of that
 * class for more information.
 * <br />
 * <br />
 * Note: Java version 1.5 contains support for enum types. These classes
 * (<code>Enumeration</code> and <code>EnumValue</code>) are provided to give
 * backwards compatibiity with previous versions of Java. 
 * <br />
 * These classes also provide the ability to define sub-lists. An
 * <code>EnumValue</code> class can extend where one
 * enumeration is an extension of another.
 * <br />
 * <br />
 * @param <V> the type of value (EnumValue) stored in this enumeration
 * @see net.lucidviews.util.EnumValue
 * @author Stephen Battey
 * @version $Revision: 1.5.2.1 $
 * @since 1.0
 */
public class Enumeration<V extends EnumValue>
{
	
	/*
	 * Local constants
	 */
	
	/** Value assigned to the <code>_superEnumeration</code> variable when there
	 *  is no super enumeration. */
	protected static final Enumeration NO_SUPER_ENUMERATION = null;
	
	
	/*
	 * Instance variables
	 */
	
	/** The super enumeration - this enumeration is a sub-list within the super
	 *  enumeration. This variable is set to <code>null</code> if there is no
	 *  super enumeration.
	 *  @since 1.2 */
	protected Enumeration<? super V> _superEnumeration;
	
	/** The values in this enumeration. */
	protected List<V> _values;
	
	
	/*
	 * Constructor(s)
	 */
	
	/**
	 * Create an enumeration, specifying the number of values so that the
	 * enumeration list can be sized at creation, thereby improving performance.
	 * <br />
	 * <br />
	 * @param initialCapacity the expected number of values in this enumeration
	 */
	public Enumeration( int initialCapacity )
	{
		this( (Enumeration<? super V>)NO_SUPER_ENUMERATION, initialCapacity );
	}
	
	/**
	 * Create an enumeration.
	 */
	public Enumeration()
	{
		this( (Enumeration<? super V>)NO_SUPER_ENUMERATION );
	}
	
	/**
	 * Create an enumeration, specifying the number of values so that the
	 * enumeration list can be sized at creation, thereby improving performance.
	 * <br />
	 * <br />
	 * @param superEnumeration the enumeration this enumeration is a sub-list of
	 * @param initialCapacity the expected number of values in this enumeration
	 * @since 1.2
	 */
	public Enumeration( Enumeration<? super V> superEnumeration, int initialCapacity )
	{
		_superEnumeration = superEnumeration;
		_values = new ArrayList<V>( initialCapacity );
	}
	
	/**
	 * Create an enumeration.
	 * <br />
	 * <br />
	 * @param superEnumeration the enumeration this enumeration is a sub-list of
	 * @since 1.2
	 */
	public Enumeration( Enumeration<? super V> superEnumeration )
	{
		_superEnumeration = superEnumeration;
		_values = new ArrayList<V>();
	}
	
	
	/*
	 * Methods
	 */
	
	/**
	 * Add a value to this enumeration.
	 * <br />
	 * An enumeration may only hold values that are of the same type - i.e.
	 * instances of the same <code>EnumValue</code> class. Attempting to add a
	 * different type of value to the enumeration will result in an exception.
	 * <br />
	 * All values must be unique. Attempting to add a duplicate value will also
	 * result in an exception.
	 * <br />
	 * <br />
	 * <code>Enumeration</code>s are not editable once they have been created.
	 * This method should only be called from within the <code>EnumValue</code>
	 * class while the list of enumeration values is being built. Therefore this
	 * method has been given package scope to prevent the list being changed
	 * after it has been created.
	 * <br />
	 * <br />
	 * @param value enumeration value to be added to this list
	 * @return the index of where the value is stored in the enumeration list
	 * @throws IllegalArgumentException invalid value for this enumeration
	 * @throws IllegalStateException invalid value for this enumeration
	 */
	int addValue( V value ) throws IllegalArgumentException, IllegalStateException
	{
		if (value == null)
			throw new IllegalArgumentException( "Cannot add a null value to an enumeration." );
		
		/* guard against duplicate elements */
		if (_values.contains( value ))
			throw new IllegalStateException( "Cannot allow duplicate enumeration value '" + value + "' of type " + value.getClass().getName() + "." );
		
		/* now store the value */
		if (!_values.add( value ))
			throw new IllegalStateException( "Addition of enumeration value '" + value + "' to enumeration of type " + value.getClass().getName() + " failed." );
		
		if (_superEnumeration == NO_SUPER_ENUMERATION)
		{
			return _values.indexOf( value );
		}
		else
		{
			/* use the index from the super enumeration
			 * that enumeration will contain more values so the index in the
			 * super list will be higher; if we used the index from the sub list
			 * it could conflict with the index of another value in the
			 * super list */
			return _superEnumeration.addValue( value );
		}
	}
	
	/**
	 * Lookup a value in this enumeration.
	 * <br />
	 * The <code>value</code> parameter can be any <code>Object</code>. The
	 * return value is always an <code>EnumValue</code> that is considered equal
	 * (using the <code>equals</code> method to compare values) to the specified
	 * <code>value</code>.
	 * <br />
	 * Thus, if the <code>Enumeration</code> contains an <code>EnumValue</code>
	 * with id 6 and string-value "yard":
	 * <br /><code>lookupValue( "yard" );</code>
	 * <br /><code>lookupValue( new Integer( 6 ) );</code>
	 * <br />
	 * will both return the <code>EnumValue</code> instance.
	 * <br />
	 * <br />
	 * @param value the value to find
	 * @return the matching enumeration value
	 * @throws NoSuchElementException the value could not be found
	 */
	public V lookupValue( Object value ) throws NoSuchElementException
	{
		Iterator<? extends V> enumValueItr = _values.iterator();
		while (enumValueItr.hasNext())
		{
			V enumValue = (V)enumValueItr.next();
			
			if (enumValue.equals( value ))
			// NB: it is critical to use 'enumValue.equals( value )' and not
			//     'value.equals( enumValue )', since the latter will almost
			//     always be false, but the EnumValue object may allow
			//     comparison with a String and find an equality
				return enumValue;
		}
		
		throw new NoSuchElementException( "No enumeration value '" + value + "' of type " + getTypeString() + " exists (enum list: " + _values + ")." );
	}
	
	/**
	 * Get the name of the type of object held in this enumeration.
	 * <br />
	 * This is an internal method used to construct debug/error messages.
	 * If handles the case where the type has not been defined
	 * (i.e. if <code>getType()</code> returns <code>null</code>).
	 * <br />
	 * <br />
	 * @return the type of object held in this enumeration
	 * @since 1.2
	 */
	protected String getTypeString()
	{
		Class<V> type = getType();
		
		return (type == null) ? "undefined" : type.getName();
			
	}
	
	
	/*
	 * Access methods
	 */
	
	/**
	 * Returns the type of enum value being stored in this enumeration.
	 * <br />
	 * This will return <code>null</code> if no values have been added to this
	 * enumeration list.
	 * <br />
	 * <br />
	 * @return the type of this enumeration
	 */
	public Class<V> getType()
	{
		if (_values.isEmpty())
		{
			return null;
		}
		
		V value = _values.get( 0 );
		//return value.getClass();
		// Logically the getClass on an object V should return Class<V>.
		// This is true if V is a class, but not so if V is a generic type.
		// So we have to make an explicit cast.
		return (Class<V>)value.getClass();
	}
	
	/**
	 * Returns the number of values in this enumeration.
	 * <br />
	 * <br />
	 * @return the number of values in this enumeration
	 */
	public int size()
	{
		return _values.size();
    }
	
	/**
	 * Tests if this enumeration has no values.
	 * <br />
	 * <br />
	 * @return <code>true</code> if no values have been defined in this
	 *         enuemration; <code>false</code> otherwise
	 */
	public boolean isEmpty()
	{
		return _values.isEmpty();
	}
	
	/**
	 * Returns an array containing all of the values in this enumeration in the
	 * correct order.
	 * <br />
	 * <br />
	 * @return an array containing all of the elements in this list in the
	 *         correct order
	 */
	public Object[] toArray()
	{
		return _values.toArray();
	}

	/**
	 * Returns an iterator of the values in this enumeration (in proper
	 * sequence).
	 * <br />
	 * <br />
	 * @return an iterator of the enumeration values (<code>EnumValue</code>s)
	 * @see net.lucidviews.util.EnumValue
	 */
	public ListIterator<? extends V> listIterator()
	{
		return _values.listIterator();
	}
	
	
	/*
	 * Methods to provide compatibility with Java5 enums.
	 */
	
	/**
	 * Get the values of this enumeration.
	 * <br />
	 * <br />
	 * @return an array of enumeration values
	 * @since 1.2
	 */
	//public V[] values()
	// NB: Java5 does not allow arrays of generics
	//    so we're forced to use the base class for the type of the return value
	public EnumValue[] values()
	{
		// The following line causes a ClassCastException
		//return (EnumValue[])_values.toArray();
		
		// so manually create the array of EnumValues
		EnumValue[] values = new EnumValue[_values.size()];
		
		int i = 0;
		Iterator<? extends V> valueItr = listIterator();
		while (valueItr.hasNext())
		{
			values[i] = valueItr.next();
			i++;
		}
		
		return values;
	}
	
	/**
	 * Lookup a value in this enumeration using its name.
	 * <br />
	 * <br />
	 * @param name the name of the enumeration value
	 * @return the enumeration value with the given name
	 * @throws IllegalArgumentException if there is no value with the specified
	 *                                  name
	 * @since 1.2
	 */
	public V valueOf( String name ) throws IllegalArgumentException
	{
		try
		{
			return lookupValue( name );
		}
		catch (NoSuchElementException e)
		{
			throw new IllegalArgumentException( "Value '" + name + "' does not exist in enumeration.", e );
		}
	}

	/**
	 * Returns the <code>EnumValue</code> of the specified enumeration type
	 * with the specified name. The name must match with the text value of
	 * the enum value. (Extraneous whitespace characters are not permitted.)
	 * <br />
	 * <br />
	 * @param enumType the <code>Class</code> object of the
	 *                 <code>EnumValue</code> from which to return a constant
	 * @param name the text value of the enum value to return
	 * @return the <code>EnumValue</code> from the specified enumeration type
	 *         with the specified name
	 * @throws IllegalArgumentException if the specified enumeration class has
	 *                                  no constant with the specified name, or
	 *                                  the specified class does not represent
	 *                                  an enum type
	 * @throws NullPointerException if <code>enumType</code> or
	 *                              <code>name</code> is <code>null</code>
	 * @since 1.2
	 */
	public static EnumValue<?> valueOf( Class<? extends EnumValue<?>> enumType, String name ) throws IllegalArgumentException, NullPointerException
	{
		return EnumValue.valueOf( enumType, name );
    }
	
	
	/*
	 * Overridden methods of Object base class
	 */
	
	/** {@inheritDoc} */
	public String toString()
	{
		StringBuffer buf = new StringBuffer();

		buf.append( "Enumeration[enumType:" );
		buf.append( getTypeString() );
		buf.append( ",values" );
		buf.append( _values );
		buf.append( "]" );

		return buf.toString();
	}
	
	/** {@inheritDoc} */
	public int hashCode()
	{
		return toString().hashCode();
	}
	
	/** {@inheritDoc} */
	public boolean equals( Object obj )
	{
		if (obj instanceof Enumeration)
		{
			Enumeration enumObj = (Enumeration)obj;

			/* all attributes must be equal (NB _enumType could be null) */
			return (_values.equals( enumObj._values ));
		}
		else
		{
			return false;
		}
	}
	
}


/*
 * Change History
 *
 * $Log: Enumeration.java,v $
 * Revision 1.5.2.1  2007/12/04 21:35:12  Steve
 * d legal disclaimer.
 * @
 * text
 * @d49 1
 * a49 1
 *  * @@version $Revision: 1.3 $
 * d56 13
 * d72 6
 * d98 23
 * d1 *
 * Revision 1.5  2007/10/05 19:25:08  Steve
 * Implemented super enumeration functionality (sub lists within lists).
 * Added methods to following the naming conventions used by the Java5 enums.
 *
 * Revision 1.4  2006/09/06 20:52:02  Steve
 * Added legal disclaimer.
 *
 * Revision 1.3  2006/08/02 19:55:59  Steve
 * Added more access methods: size, isEmpty, toArray, listIterator.
 *
 * Revision 1.2  2006/04/27 21:05:14  Steve
 * Added getType method.
 *
 * Revision 1.1  2006/04/17 15:33:38  Steve
 * Initial revision.
 *
 */
