/*
 * Properties.java
 * Part of the B'am library - a collection of general-purpose utility classes.
 * 
 * Copyright (C)2006 Stephen Battey
 * Contact: resource@lucidviews.net
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

package net.lucidviews.util;

import net.lucidviews.util.io.Streams;
import net.lucidviews.util.text.ParseValueException;
import net.lucidviews.util.text.Strings;

import java.awt.Color;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.IOException;
import java.net.URL;


/**
 * An extension to the <code>java.util.Properties</code> class, which parses and
 * validates non-<code>String</code> values.
 * <br />
 * <br />
 * A <code>net.lucidviews.util.Properties</code> class can be instantiated in a
 * number of ways:
 * <ul>
 *   <li>
 *     Using the default constructor to create an empty table of values.
 *   </li>
 *   <li>
 *     With a <code>File</code> to load an initial set of property values from a
 *     <code>.properties</code> file. (See the file format described in the
 *     <code>java.util.Properties</code> JavaDoc.)
 *   </li>
 *   <li>
 *     With a <code>java.util.Properties</code> object. Effectively using the
 *     <code>net.lucidviews.util.Properties</code> class to wrapper the basic
 *     properties table and provide more advanced access to the values therein. 
 *   </li>
 * </ul>
 * <br />
 * The <code>net.lucidviews.util.Properties</code> class provides access methods
 * to set and get values of various types, including:
 * <ul>
 *   <li><code>java.lang.Boolean</code> (and <code>boolean</code>)</li>
 *   <li><code>java.lang.Class</code></li>
 *   <li><code>java.awt.Color</code></li>
 *   <li><code>java.lang.Double</code> (and <code>double</code>)</li>
 *   <li><code>net.lucidviews.util.EnumValue</code></li>
 *   <li><code>java.io.File</code></li>
 *   <li><code>java.lang.Long</code> (and <code>long</code>)</li>
 * </ul>
 * Each getter method is overloaded to allow the caller the option of providing
 * default values, giving total flexibility in error handling.
 * <br />
 * <br />
 * The 
 * <br />
 * <br />
 * @see net.lucidviews.util.PropertyException
 * @see java.util.Properties
 * @author Stephen Battey
 * @version $Revision: 1.6 $
 * @since 1.0
 */
public class Properties extends java.util.Properties
{
	
	/*
	 * Exported constants
	 */
	
	/** Special default value that ensures a property has been defined for a
	 *  given Boolean value.
	 *  @see #getBooleanProperty(String,Boolean,Boolean) */
	public static final Boolean BOOLEAN_PARAM_MUST_EXIST = new Boolean( false );
	
	/** Special default value that ensures a property has been defined for a
	 *  given integer value.
	 *  @see #getLongProperty(String,Long,Long) */
	public static final Long LONG_PARAM_MUST_EXIST = new Long( 0 );
	
	/** Special default value that ensures a property has been defined for a
	 *  given floating-point value.
	 *  @see #getDoubleProperty(String,Double,Double) */
	public static final Double DOUBLE_PARAM_MUST_EXIST = new Double( 0 );
	
	/** Special default value that ensures a property has been defined for a
	 *  given file location.
	 *  @see #getFileProperty(String,File,File) */
	public static final File FILE_PARAM_MUST_EXIST = new File( "" );
	
	/** Special default value that ensures a property has been defined for a
	 *  given folder location.
	 *  @see #getFolderProperty(String,File,File) */
	public static final File FOLDER_PARAM_MUST_EXIST = new File( "" );
	
	/** Special default value that ensures a property has been defined for a
	 *  given class reference.
	 *  @see #getClassProperty(String,Class,Class) */
	public static final Class CLASS_PARAM_MUST_EXIST = ClassParamMustExist.class;
	
	/** Special default value that ensures a property has been defined for a
	 *  given enumeration value.
	 *  @see #getEnumProperty(String,Enumeration,EnumValue) */
	public static final EnumValue ENUM_PARAM_MUST_EXIST = new EnumParamMustExist();
	
	/** Special default value that ensures a property has been defined for a
	 *  given colour value.
	 *  @see #getColorProperty(String,Color,Color) */
	public static final Color COLOR_PARAM_MUST_EXIST = new Color( 0 );
	
	
	/*
	 * Mirrored constants from Strings class
	 */
	
	/** Special default value that ensures a valid Boolean value is parsed
	 *  (i.e. to be used when no default value is applicable).
	 *  @see #getBooleanProperty(String,Boolean,Boolean) */
	public static final Boolean BOOLEAN_MUST_BE_VALID = Strings.BOOLEAN_MUST_BE_VALID;
	
	/** Special default value that ensures a valid integer value is parsed
	 *  (i.e. to be used when no default value is applicable).
	 *  @see #getLongProperty(String,Long,Long) */
	public static final Long LONG_MUST_BE_VALID = Strings.LONG_MUST_BE_VALID;
	
	/** Special default value that ensures a valid floating-point value is
	 *  parsed (i.e. to be used when no default value is applicable).
	 *  @see #getDoubleProperty(String,Double,Double) */
	public static final Double DOUBLE_MUST_BE_VALID = Strings.DOUBLE_MUST_BE_VALID;
	
	/** Special default value that ensures a valid File is parsed and references
	 *  a file that already exists on the file system.
	 *  @see #getFileProperty(String,File,File) */
	public static final File FILE_MUST_EXIST = Strings.FILE_MUST_EXIST;
	
	/** Special default value that ensures a valid File is parsed and references
	 *  an existing file or a location where a new file can be created.
	 *  @see #getFileProperty(String,File,File) */
	public static final File ENSURE_FILE_EXISTS = Strings.ENSURE_FILE_EXISTS;
	
	/** Special default value that ensures a valid File is parsed and references
	 *  a folder that already exists on the file system.
	 *  @see #getFolderProperty(String,File,File) */
	public static final File FOLDER_MUST_EXIST = Strings.FOLDER_MUST_EXIST;
	
	/** Special default value that ensures a valid File is parsed and references
	 *  an existing folder or a location where a new folder can be created.
	 *  @see #getFolderProperty(String,File,File) */
	public static final File ENSURE_FOLDER_EXISTS = Strings.ENSURE_FOLDER_EXISTS;
	
	/** Special default value that ensures a valid File is parsed and references
	 *  a location within an existing folder on the file system.
	 *  @see #getFileProperty(String,File,File)
	 *  @see #getFolderProperty(String,File,File) */
	public static final File PARENT_FOLDER_MUST_EXIST = Strings.PARENT_FOLDER_MUST_EXIST;
	
	/** Special default value that ensures a valid File is parsed and references
	 *  a location within an existing folder or in a folder that can be created.
	 *  @see #getFileProperty(String,File,File)
	 *  @see #getFolderProperty(String,File,File) */
	public static final File ENSURE_PARENT_FOLDER_EXISTS = Strings.ENSURE_PARENT_FOLDER_EXISTS;
	
	/** Special default value that ensures a valid Class is parsed
	 *  (i.e. to be used when no default value is applicable).
	 *  @see #getClassProperty(String,Class,Class) */
	public static final Class CLASS_MUST_BE_VALID = Strings.CLASS_MUST_BE_VALID;
	
	/** Special default value that ensures a valid Enumeration value is parsed
	 *  (i.e. to be used when no default value is applicable).
	 *  @see #getEnumProperty(String, Enumeration,EnumValue,EnumValue) */
	public static final EnumValue ENUM_VALUE_MUST_BE_VALID = Strings.ENUM_VALUE_MUST_BE_VALID;
	
	/** Special default value that ensures a valid Color is parsed
	 *  (i.e. to be used when no default value is applicable).
	 *  @see #getColorProperty(String,Color,Color) */
	public static final Color COLOR_MUST_BE_VALID = Strings.COLOR_MUST_BE_VALID;
	
	
	/*
	 * Local constants
	 */
	
	/** A default source description. */
	protected static final String UNDEFINED_SOURCE = "<undefined source>";
	
	/** Version of class serialisation. */
	protected static final long serialVersionUID = 0l;
	
	
	/*
	 * Instance variables
	 */
	
	/** A description of where the properties were defined. */
	protected String _sourceDescription = UNDEFINED_SOURCE;
	
	/** The file used to persist these property values. */
	protected File _sourceFile;
	
	
	/*
	 * Constructors
	 */
	
	/**
	 * Set-up an empty table of property values.
	 * <br />
	 * Clients are encouraged to use the <code>Properties( String )</code>
	 * constructor instead and specify a description of the properties table, so
	 * that messages relating to property values can be put in context
	 * <br />
	 * e.g. "The database.url property in the <b>myappn.properties file</b>"
	 */
	public Properties()
	{
		this( UNDEFINED_SOURCE );
	}

	/**
	 * Set-up an empty table of property values, with a description of the
	 * source of the property values.
	 * <br />
	 * The description will be used to give messages relating to property values
	 * some context
	 * <br />
	 * e.g. "The database.url property in the <b>myappn.properties file</b>"
	 * <br />
	 * @param sourceDescription a description of where the property values will
	 *                          be loaded from
	 */
	public Properties( String sourceDescription )
	{
		super();

		_sourceDescription = sourceDescription;
	}
	
	/**
	 * Set-up a table of property values loaded from a given source file.
	 * <br />
	 * The file will also be taken as the source of the properties - meaning a
	 * link will be established between the properties table and the content of
	 * the file. If the <code>store</code> or <code>load</code> methods are
	 * called it will be assumed the properties are to be stored/loaded from
	 * this source file, unless a different file is stated.
	 * <br />
	 * <br />
	 * @param sourceFile the file where the property values are to be loaded
	 *                   from and stored to
	 * @throws IOException an error prevented the values being read from the
	 *                     specified file
	 * @see #load()
	 * @see #store()
	 * @see #setSourceFile
	 */
	public Properties( File sourceFile ) throws IOException
	{
		super();
		
		setSourceFile( sourceFile );
		try
		{
			load();
		}
		catch (FileNotFoundException e)
		{
			// the client may be creating a new set of properties - so if the
			// source file does not exist allow a blank Properties table to be
			// created
		}
	}

	/**
	 * Set-up a table of property values copying values from another
	 * <code>Properties</code> table.
	 * <br />
	 * Clients are encouraged to use the
	 * <code>Properties( Properties, String )</code> constructor instead and
	 * specify a description of the properties table, so that messages relating
	 * to property values can be put in context
	 * <br />
	 * e.g. "The database.url property in the <b>myappn.properties file</b>"
	 * <br />
	 * @param properties initial values to be loaded into this properties table
	 */
	public Properties( java.util.Properties properties )
	{
		this( properties, UNDEFINED_SOURCE );
	}

	/**
	 * Set-up a table of property values copying values from another
	 * <code>Properties</code> table, with a description of the
	 * source of the property values.
	 * <br />
	 * The description will be used to give messages relating to property values
	 * some context
	 * <br />
	 * e.g. "The database.url property in the <b>myappn.properties file</b>"
	 * <br />
	 * @param properties initial values to be loaded into this properties table
	 * @param sourceDescription a description of where the property values will
	 *                          be loaded from
	 */
	public Properties( java.util.Properties properties, String sourceDescription )
	{
		super( properties );

		_sourceDescription = sourceDescription;
	}


	/*
	 * Methods
	 */
	
	/**
	 * Set the file that is used to persist these properties.
	 * <br />
	 * The given file will be implicitly linked to this table of properties
	 * table. If the <code>store</code> or <code>load</code> methods are
	 * called it will be assumed the properties are to be stored/loaded from
	 * this source file, unless a different file is stated.
	 * <br />
	 * <br />
	 * @param sourceFile the file where the property values are to be loaded
	 *                   and stored
	 * @see #store()
	 * @see #load()
	 */
	public void setSourceFile( File sourceFile )
	{
		_sourceFile = sourceFile;
	}

	/**
	 * Load property values from the source file.
	 * <br />
	 * The properties will be added to those already contained in this
	 * <code>Properties</code> table (to remove values already loaded, use the
	 * <code>clear</code> method).
	 * <br />
	 * If a source file has not been set, either by using the
	 * <code>Properties( File )</code> constructor or by calling the
	 * <code>setSourceFile</code> method, then an exception will be thrown.
	 * <br />
	 * The contents of the file must meet the specification described by the
	 * <code>load</code> method of the <code>java.util.Properties</code> class.
	 * <br />
	 * <br />
	 * @throws FileNotFoundException no source file has been defined for this
	 *                               <code>Properties</code> table or the
	 *                               specified source file does not exist
	 * @throws IOException an error prevented the values being read from the
	 *                     specified file
	 * @see java.util.Properties#load
	 */
	public void load() throws FileNotFoundException, IOException
	{
		if (_sourceFile == null)
			throw new FileNotFoundException( "No source file has been assigned to the Properties table." );
		
		load( _sourceFile );
	}

	/**
	 * Load property values from a specified file.
	 * <br />
	 * The properties will be added to those already contained in this
	 * <code>Properties</code> table (to remove values already loaded, use the
	 * <code>clear</code> method).
	 * <br />
	 * The contents of the file must meet the specification described by the
	 * <code>load</code> method of the <code>java.util.Properties</code> class.
	 * <br />
	 * <br />
	 * @param inFile the file where the property values are defined
	 * @throws FileNotFoundException the specified file does not exist
	 * @throws IOException an error prevented the values being read from the
	 *                     specified file
	 * @see java.util.Properties#load
	 */
	public void load( File inFile ) throws FileNotFoundException, IOException
	{
		FileInputStream fileStream = null;
		try
		{
			fileStream = new FileInputStream( inFile );
			super.load( fileStream );
		}
		finally
		{
			// close the input stream, even if an error occurs
			Streams.close( fileStream );
		}
	}

	/**
	 * Load property values from a specified URL.
	 * <br />
	 * The properties will be added to those already contained in this
	 * <code>Properties</code> table (to remove values already loaded, use the
	 * <code>clear</code> method).
	 * <br />
	 * The contents of the file must meet the specification described by the
	 * <code>load</code> method of the <code>java.util.Properties</code> class.
	 * <br />
	 * <br />
	 * @param inUrl the location of the resource where the property values are
	 *            defined
	 * @throws IOException an error prevented the values being read from the
	 *                     specified resource
	 * @see java.util.Properties#load
	 */
	public void load( URL inUrl ) throws IOException
	{
		InputStream urlStream = null;
		try
		{
			urlStream = inUrl.openStream();
			super.load( urlStream );
		}
		finally
		{
			// close the input stream, even if an error occurs
			Streams.close( urlStream );
		}
	}

	/**
	 * Store the property values into the source file.
	 * <br />
	 * Any content in the file will be overwritten.
	 * <br />
	 * <br />
	 * @throws IOException error writing the property values to the source file
	 * @see #setSourceFile
	 */
	public void store() throws IOException
	{
		store( "" );
	}

	/**
	 * Store the property values into the source file.
	 * <br />
	 * Any content in the file will be overwritten.
	 * <br />
	 * <br />
	 * @param header a description of the property list
	 * @throws IOException error writing the property values to the source file
	 * @see #setSourceFile
	 */
	public void store( String header ) throws IOException
	{
		store( _sourceFile, header );
	}

	/**
	 * Store the property values into a file at the given location.
	 * <br />
	 * Any content in the file will be overwritten.
	 * <br />
	 * <br />
	 * @param outFile the file where the property values are to be stored
	 * @param header a description of the property list
	 * @throws IOException error writing the property values to the specified
	 *                     file
	 */
	public void store( File outFile, String header ) throws IOException
	{
		FileOutputStream outStream = null;
		try
		{
			outStream = new FileOutputStream( outFile, false );
			super.store( outStream, header );
			outStream.flush();
		}
		finally
		{
			// close the output stream, even if an error occurs
			outStream.close();
		}
	}


	/*
	 * Get methods
	 */

	/**** ****    STRING PROPERTY    **** ****/

	/**
	 * Retrieve a String value from the property with given key.
	 * <br />
	 * If the property with given key does not exist an exception will be
	 * thrown.
	 * <br />
	 * The return value is guaranteed non-<code>null</code>. Thus, the
	 * following line of code will never cause a
	 * <code>NullPointerException</code>:
	 * <code>
	 * <br />int strLen = myproperties.getStringProperty( <i>key</i> ).length();
	 * </code>
	 * <br />
	 * Although it may throw a <code>PropertyException</code> if the property
	 * does not exist.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @return the text value of the given property
	 * @throws PropertyException the property does not exist
	 */
	public String getStringProperty( String key ) throws PropertyException
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
			throw new PropertyException( _sourceDescription, key );

		return stringValue;
	}

	/**
	 * Retrieve a String value from the property with given key, returning the
	 * default value if the property is missing.
	 * <br />
	 * If the property with given key does not exist then the default value will
	 * be returned.
	 * <br />
	 * The return value is guaranteed non-<code>null</code> provided the
	 * <code>defaultValue</code> is not <code>null</code>. Thus, the following
	 * line of code will never cause a <code>NullPointerException</code>:
	 * <code>
	 * <br />int strLen = myproperties.getStringProperty( <i>key</i>, <i>default</i> ).length();
	 * </code>
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @param defaultValue the value to be used if the property is missing
	 * @return the text value of the given property
	 */
	public String getStringProperty( String key, String defaultValue )
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
			return defaultValue;

		return stringValue;
	}
	
	/**
	 * Assign a text value to a given property.
	 * <br />
	 * <br />
	 * @param key the name of the property to be stored
	 * @param value the new value of the property
	 */
	public void setStringProperty( String key, String value )
	{
		setProperty( key, value );
	}

	/**** ****    BOOLEAN PROPERTY    **** ****/

	/**
	 * Parse a Boolean value from the property with given key.
	 * <br />
	 * If the property with given key does not exist an exception will be
	 * thrown.
	 * <br />
	 * If the value of the property with given key does not represent a valid
	 * Boolean value then an exception will be thrown.
	 * <br />
	 * The return value is guaranteed non-<code>null</code>. Thus, the
	 * following line of code will never cause a
	 * <code>NullPointerException</code>:
	 * <code>
	 * <br />boolean value = myproperties.getBooleanProperty( <i>key</i> ).booleanValue();
	 * </code>
	 * <br />
	 * Although it may throw a <code>PropertyException</code> if the property
	 * value is invalid.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @return the Boolean value of the given property
	 * @throws PropertyException the property does not exist or is invalid
	 */
	public Boolean getBooleanProperty( String key ) throws PropertyException
	{
		return getBooleanProperty( key, BOOLEAN_PARAM_MUST_EXIST, BOOLEAN_MUST_BE_VALID );
	}

	/**
	 * Parse a Boolean value from the property with given key, returning one of
	 * the default values if the property is missing or invalid.
	 * <br />
	 * If the property with given key does not exist and the
	 * <code>defaultIfMissing</code> parameter is set to
	 * <code>Properties.BOOLEAN_PARAM_MUST_EXIST</code> then an exception will
	 * be thrown. Otherwise the <code>defaultIfMissing</code> value is returned.
	 * <br />
	 * If the value of the property with given key does not represent a valid
	 * Boolean value and the <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.BOOLEAN_MUST_BE_VALID</code> then an exception will
	 * be thrown. Otherwise the <code>defaultIfInvalid</code> value is returned.
	 * <br />
	 * The return value is guaranteed non-<code>null</code> provided neither of
	 * the <code>defaultValue</code>s are <code>null</code>. Thus, the
	 * following line of code will never cause a
	 * <code>NullPointerException</code>:
	 * <code>
	 * <br />boolean value = myproperties.getBooleanProperty( <i>key</i>, <i>default</i>, <i>default</i> ).booleanValue();
	 * </code>
	 * <br />
	 * Although it may throw a <code>PropertyException</code> if the property
	 * value is invalid.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @param defaultIfMissing the Boolean value to use if the property is
	 *                         missing (use
	 *                         <code>Properties.BOOLEAN_PARAM_MUST_EXIST</code>
	 *                         to cause an exception in such a situation)
	 * @param defaultIfInvalid the Boolean value to use if the property does not
	 *                         specify a valid Boolean value (use
	 *                         <code>Properties.BOOLEAN_MUST_BE_VALID</code> to
	 *                         cause an exception in such a situation)
	 * @return the Boolean value of the given property
	 * @throws PropertyException the property does not exist or is invalid
	 */
	public Boolean getBooleanProperty( String key, Boolean defaultIfMissing, Boolean defaultIfInvalid ) throws PropertyException
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
		{
			if (defaultIfMissing == BOOLEAN_PARAM_MUST_EXIST)
				throw new PropertyException( _sourceDescription, key );
			else
				return defaultIfMissing;
		}

		try
		{
			return Strings.parseBoolean( defaultIfInvalid, stringValue );
		}
		catch (ParseValueException e)
		{
			throw new PropertyException( _sourceDescription, key, stringValue, e );
		}
	}

	/**
	 * Parse a Boolean value from the property with given key, returning the
	 * default value if the property is missing or invalid.
	 * <br />
	 * If the property with given key does not exist or does not represent a
	 * valid Boolean value then the default value will be returned.
	 * <br />
	 * The return value is guaranteed non-<code>null</code> provided the
	 * <code>defaultValue</code> is not <code>null</code>. Thus, the following
	 * line of code will never cause a <code>NullPointerException</code>:
	 * <code>
	 * <br />boolean value = myproperties.getBooleanProperty( <i>key</i>, <i>default</i> ).booleanValue();
	 * </code>
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @param defaultValue the value to be used if the property is missing or
	 *                     does not represent a Boolean value
	 * @return the Boolean value of the given property
	 */
	public Boolean getBooleanProperty( String key, Boolean defaultValue )
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
			return defaultValue;

		return Strings.parseBoolean( stringValue, defaultValue );
	}

	/**
	 * Parse a Boolean value from the property with given key, returning the
	 * default value if the property is missing or invalid.
	 * <br />
	 * If the property with given key does not exist or does not represent a
	 * valid Boolean value then the default value will be returned.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @param defaultValue the value to be used if the property is missing or
	 *                     does not represent a Boolean value
	 * @return the Boolean value of the given property
	 */
	public boolean getBooleanProperty( String key, boolean defaultValue )
	{
		return getBooleanProperty( key, defaultValue, defaultValue );
	}

	/**
	 * Parse a Boolean value from the property with given key, returning one of
	 * the default values if the property is missing or invalid.
	 * <br />
	 * If the property with given key does not exist, the
	 * <code>defaultIfMissing</code> value will be returned.
	 * <br />
	 * If the value of the property with given key does not represent a valid
	 * Boolean value then the <code>defaultIfInvalid</code> value is returned.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @param defaultIfMissing the value to be used if the property is missing
	 * @param defaultIfInvalid the value to be used if the property does not
	 *                         represent a Boolean value
	 * @return the Boolean value of the given property
	 */
	public boolean getBooleanProperty( String key, boolean defaultIfMissing, boolean defaultIfInvalid )
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
			return defaultIfMissing;

		return Strings.parseBoolean( stringValue, defaultIfInvalid );
	}

	/**
	 * Assign a Boolean value to a given property.
	 * <br />
	 * <br />
	 * @param key the name of the property to be stored
	 * @param value the new value of the property
	 */
	public void setBooleanProperty( String key, Boolean value )
	{
		setBooleanProperty( key, value.booleanValue() );
	}
	
	/**
	 * Assign a Boolean value to a given property.
	 * <br />
	 * <br />
	 * @param key the name of the property to be stored
	 * @param value the new value of the property
	 */
	public void setBooleanProperty( String key, boolean value )
	{
		if (value)
			setProperty( key, "true" );
		else
			setProperty( key, "false" );
	}

	/**** ****    GET NUMERIC, INTEGER PROPERTY    **** ****/

	/**
	 * Parse a numeric, integer value from the property with given key.
	 * <br />
	 * If the property with given key does not exist an exception will be
	 * thrown.
	 * <br />
	 * If the value of the property with given key does not represent a valid
	 * integer value then an exception will be thrown.
	 * <br />
	 * The return value is guaranteed non-<code>null</code>. Thus, the
	 * following lines of code will never cause a
	 * <code>NullPointerException</code>:
	 * <code>
	 * <br />long value = myproperties.getLongProperty( <i>key</i> ).longValue();
	 * <br />int value = myproperties.getLongProperty( <i>key</i> ).intValue();
	 * </code>
	 * <br />
	 * Although they may throw a <code>PropertyException</code> if the
	 * property value is invalid.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @return the numeric, integer value of the given property
	 * @throws PropertyException the property does not exist or is invalid
	 */
	public Long getLongProperty( String key ) throws PropertyException
	{
		return getLongProperty( key, LONG_PARAM_MUST_EXIST, LONG_MUST_BE_VALID );
	}

	/**
	 * Parse a numeric, integer value from the property with given key,
	 * returning one of the default values if the property is missing or
	 * invalid.
	 * <br />
	 * If the property with given key does not exist and the
	 * <code>defaultIfMissing</code> parameter is set to
	 * <code>Properties.LONG_PARAM_MUST_EXIST</code> then an exception will
	 * be thrown. Otherwise the <code>defaultIfMissing</code> value is returned.
	 * <br />
	 * If the value of the property with given key does not represent a valid
	 * numeric, integer value and the <code>defaultIfInvalid</code> parameter is
	 * set to <code>Properties.LONG_MUST_BE_VALID</code> then an exception will
	 * be thrown. Otherwise the <code>defaultIfInvalid</code> value is returned.
	 * <br />
	 * The return value is guaranteed non-<code>null</code> provided neither of
	 * the <code>defaultValue</code>s are <code>null</code>. Thus, the
	 * following lines of code will never cause a
	 * <code>NullPointerException</code>:
	 * <code>
	 * <br />long value = myproperties.getLongProperty( <i>key</i>, <i>default</i>, <i>default</i> ).longValue();
	 * <br />int value = myproperties.getLongProperty( <i>key</i>, <i>default</i>, <i>default</i> ).intValue();
	 * </code>
	 * <br />
	 * Although it may throw a <code>PropertyException</code> if the property
	 * value is invalid.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @param defaultIfMissing the number to use if the property is missing (use
	 *                         <code>Properties.LONG_PARAM_MUST_EXIST</code> to
	 *                         cause an exception in such a situation)
	 * @param defaultIfInvalid the number to use if the property does not
	 *                         specify a valid integer value (use
	 *                         <code>Properties.LONG_MUST_BE_VALID</code> to
	 *                         cause an exception in such a situation)
	 * @return the numeric, integer value of the given property
	 * @throws PropertyException the property does not exist or is invalid
	 */
	public Long getLongProperty( String key, Long defaultIfMissing, Long defaultIfInvalid ) throws PropertyException
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
		{
			if (defaultIfMissing == LONG_PARAM_MUST_EXIST)
				throw new PropertyException( _sourceDescription, key );
			else
				return defaultIfMissing;
		}

		try
		{
			return Strings.parseLong( defaultIfInvalid, stringValue );
		}
		catch (ParseValueException e)
		{
			throw new PropertyException( _sourceDescription, key, stringValue, e );
		}
	}

	/**
	 * Parse a numeric, integer value from the property with given key,
	 * returning the default value if the property is missing or invalid.
	 * <br />
	 * If the property with given key does not exist or does not represent a
	 * valid numeric, integer value then the default value will be returned.
	 * <br />
	 * The return value is guaranteed non-<code>null</code> provided the
	 * <code>defaultValue</code> is not <code>null</code>. Thus, the following
	 * lines of code will never cause a <code>NullPointerException</code>:
	 * <code>
	 * <br />long value = myproperties.getLongProperty( <i>key</i>, <i>default</i> ).longValue();
	 * <br />int value = myproperties.getLongProperty( <i>key</i>, <i>default</i> ).intValue();
	 * </code>
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @param defaultValue the value to be used if the property is missing or
	 *                     does not represent a numeric, integer value
	 * @return the numeric, integer value of the given property
	 */
	public Long getLongProperty( String key, Long defaultValue )
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
			return defaultValue;

		return Strings.parseLong( stringValue, defaultValue );
	}

	/**
	 * Parse a numeric, integer value from the property with given key,
	 * returning the default value if the property is missing or invalid.
	 * <br />
	 * If the property with given key does not exist or does not represent a
	 * valid numeric, integer value then the default value will be returned.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @param defaultValue the value to be used if the property is missing or
	 *                     does not represent a numeric, integer value
	 * @return the numeric, integer value of the given property
	 */
	public long getLongProperty( String key, long defaultValue )
	{
		return getLongProperty( key, defaultValue, defaultValue );
	}

	/**
	 * Parse a numeric, integer value from the property with given key,
	 * returning one of the default values if the property is missing or
	 * invalid.
	 * <br />
	 * If the property with given key does not exist, the
	 * <code>defaultIfMissing</code> value will be returned.
	 * <br />
	 * If the value of the property with given key does not represent a valid
	 * numeric, integer value then the <code>defaultIfInvalid</code> value is
	 * returned.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @param defaultIfMissing the value to be used if the property is missing
	 * @param defaultIfInvalid the value to be used if the property does not
	 *                         represent a numeric, integer value
	 * @return the numeric, integer value of the given property
	 */
	public long getLongProperty( String key, long defaultIfMissing, long defaultIfInvalid )
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
			return defaultIfMissing;

		return Strings.parseLong( stringValue, defaultIfInvalid );
	}

	/**
	 * Assign an integer value to a given property.
	 * <br />
	 * <br />
	 * @param key the name of the property to be stored
	 * @param value the new value of the property
	 */
	public void setLongProperty( String key, Long value )
	{
		setProperty( key, value.toString() );
	}
	
	/**
	 * Assign an integer value to a given property.
	 * <br />
	 * <br />
	 * @param key the name of the property to be stored
	 * @param value the new value of the property
	 */
	public void setLongProperty( String key, long value )
	{
		setProperty( key, String.valueOf( value ) );
	}

	/**** ****    GET NUMERIC, FLOATING-POINT PROPERTY    **** ****/

	/**
	 * Parse a numeric, floating-point value from the property with given key.
	 * <br />
	 * If the property with given key does not exist an exception will be
	 * thrown.
	 * <br />
	 * If the value of the property with given key does not represent a valid
	 * floating-point value then an exception will be thrown.
	 * <br />
	 * The return value is guaranteed non-<code>null</code>. Thus, the
	 * following lines of code will never cause a
	 * <code>NullPointerException</code>:
	 * <code>
	 * <br />double value = myproperties.getDoubleProperty( <i>key</i> ).doubleValue();
	 * <br />float value = myproperties.getDoubleProperty( <i>key</i> ).floatValue();
	 * </code>
	 * <br />
	 * Although they may throw a <code>PropertyException</code> if the property
	 * value is invalid.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @return the numeric, floating-point value of the given property
	 * @throws PropertyException the property does not exist or is invalid
	 */
	public Double getDoubleProperty( String key ) throws PropertyException
	{
		return getDoubleProperty( key, DOUBLE_PARAM_MUST_EXIST, DOUBLE_MUST_BE_VALID );
	}

	/**
	 * Parse a numeric, floating-point value from the property with given key,
	 * returning one of the default values if the property is missing or
	 * invalid.
	 * <br />
	 * If the property with given key does not exist and the
	 * <code>defaultIfMissing</code> parameter is set to
	 * <code>Properties.DOUBLE_PARAM_MUST_EXIST</code> then an exception will
	 * be thrown. Otherwise the <code>defaultIfMissing</code> value is returned.
	 * <br />
	 * If the value of the property with given key does not represent a valid
	 * numeric, floating-point value and the <code>defaultIfInvalid</code>
	 * parameter is set to <code>Properties.DOUBLE_MUST_BE_VALID</code> then an
	 * exception will be thrown. Otherwise the <code>defaultIfInvalid</code>
	 * value is returned.
	 * <br />
	 * The return value is guaranteed non-<code>null</code> provided neither of
	 * the <code>defaultValue</code>s are <code>null</code>. Thus, the
	 * following lines of code will never cause a
	 * <code>NullPointerException</code>:
	 * <code>
	 * <br />double value = myproperties.getDoubleProperty( <i>key</i>, <i>default</i>, <i>default</i> ).doubleValue();
	 * <br />float value = myproperties.getDoubleProperty( <i>key</i>, <i>default</i>, <i>default</i> ).floatValue();
	 * </code>
	 * <br />
	 * Although it may throw a <code>PropertyException</code> if the property
	 * value is invalid.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @param defaultIfMissing the number to use if the property is missing (use
	 *                         <code>Properties.DOUBLE_PARAM_MUST_EXIST</code>
	 *                         to cause an exception in such a situation)
	 * @param defaultIfInvalid the number to use if the property does not
	 *                         specify a valid floating-point value (use
	 *                         <code>Properties.DOUBLE_MUST_BE_VALID</code> to
	 *                         cause an exception in such a situation)
	 * @return the numeric, floating-point value of the given property
	 * @throws PropertyException the property does not exist or is invalid
	 */
	public Double getDoubleProperty( String key, Double defaultIfMissing, Double defaultIfInvalid ) throws PropertyException
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
		{
			if (defaultIfMissing == DOUBLE_PARAM_MUST_EXIST)
				throw new PropertyException( _sourceDescription, key );
			else
				return defaultIfMissing;
		}

		try
		{
			return Strings.parseDouble( defaultIfInvalid, stringValue );
		}
		catch (ParseValueException e)
		{
			throw new PropertyException( _sourceDescription, key, stringValue, e );
		}
	}

	/**
	 * Parse a numeric, floating-point value from the property with given key,
	 * returning the default value if the property is missing or invalid.
	 * <br />
	 * If the property with given key does not exist or does not represent a
	 * valid numeric, floating-point value then the default value will be
	 * returned.
	 * <br />
	 * The return value is guaranteed non-<code>null</code> provided the
	 * <code>defaultValue</code> is not <code>null</code>. Thus, the following
	 * lines of code will never cause a <code>NullPointerException</code>:
	 * <code>
	 * <br />double value = myproperties.getDoubleProperty( <i>key</i>, <i>default</i> ).doubleValue();
	 * <br />float value = myproperties.getDoubleProperty( <i>key</i>, <i>default</i> ).floatValue();
	 * </code>
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @param defaultValue the value to be used if the property is missing or
	 *                     does not represent a numeric, floating-point value
	 * @return the numeric, floating-point value of the given property
	 */
	public Double getDoubleProperty( String key, Double defaultValue )
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
			return defaultValue;

		return Strings.parseDouble( stringValue, defaultValue );
	}

	/**
	 * Parse a numeric, floating-point value from the property with given key,
	 * returning the default value if the property is missing or invalid.
	 * <br />
	 * If the property with given key does not exist or does not represent a
	 * valid numeric, floating-point value then the default value will be
	 * returned.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @param defaultValue the value to be used if the property is missing or
	 *                     does not represent a numeric, floating-point value
	 * @return the numeric, floating-point value of the given property
	 */
	public double getDoubleProperty( String key, double defaultValue )
	{
		return getDoubleProperty( key, defaultValue, defaultValue );
	}

	/**
	 * Parse a numeric, floating-point value from the property with given key,
	 * returning one of the default values if the property is missing or
	 * invalid.
	 * <br />
	 * If the property with given key does not exist, the
	 * <code>defaultIfMissing</code> value will be returned.
	 * <br />
	 * If the value of the property with given key does not represent a valid
	 * numeric, floating-point value then the <code>defaultIfInvalid</code>
	 * value is returned.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @param defaultIfMissing the value to be used if the property is missing
	 * @param defaultIfInvalid the value to be used if the property does not
	 *                         represent a numeric, floating-point value
	 * @return the numeric, floating-point value of the given property
	 */
	public double getDoubleProperty( String key, double defaultIfMissing, double defaultIfInvalid )
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
			return defaultIfMissing;

		return Strings.parseDouble( stringValue, defaultIfInvalid );
	}

	/**
	 * Assign a (floating-point) numeric value to a given property.
	 * <br />
	 * <br />
	 * @param key the name of the property to be stored
	 * @param value the new value of the property
	 */
	public void setDoubleProperty( String key, Double value )
	{
		setProperty( key, value.toString() );
	}
	
	/**
	 * Assign a (floating-point) numeric value to a given property.
	 * <br />
	 * <br />
	 * @param key the name of the property to be stored
	 * @param value the new value of the property
	 */
	public void setDoubleProperty( String key, double value )
	{
		setProperty( key, String.valueOf( value ) );
	}

	/**** ****    GET FILE PROPERTY    **** ****/

	/**
	 * Parse a file reference from the property with given key.
	 * <br />
	 * If the property with given key does not exist an exception will be
	 * thrown.
	 * <br />
	 * If the value of the property with given key does not specify a valid file
	 * path then an exception will be thrown.
	 * <br />
	 * The return value is guaranteed non-<code>null</code>.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @return file reference of the location specified by the given property
	 * @throws PropertyException the property does not exist or is invalid
	 */
	public File getFileProperty( String key ) throws PropertyException
	{
		return getFileProperty( key, FILE_PARAM_MUST_EXIST, FILE_MUST_EXIST );
	}

	/**
	 * Parse a file reference from the property with given key, returning one of
	 * the default values if the property is missing or invalid.
	 * <br />
	 * If the property with given key does not exist and the
	 * <code>defaultIfMissing</code> parameter is set to
	 * <code>Properties.FILE_PARAM_MUST_EXIST</code> then an exception will
	 * be thrown. Otherwise the <code>defaultIfMissing</code> value is returned.
	 * <br />
	 * If the value of the property with given key does not specify a valid file
	 * path and the <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.FILE_MUST_EXIST</code>, an exception will be thrown.
	 * <br />
	 * If the value of the property with given key does not specify a valid file
	 * path and the <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.ENSURE_FILE_EXISTS</code>, a new, empty file will be
	 * created and a reference to that new file returned.
	 * <br />
	 * If the value of the property with given key specifies a path to a file in
	 * a folder that does not exist and the <code>defaultIfInvalid</code>
	 * parameter is set to <code>Properties.PARENT_FOLDER_MUST_EXIST</code>, an
	 * exception will be thrown.
	 * <br />
	 * If the value of the property with given key specifies a path to a file in
	 * a folder that does not exist and the <code>defaultIfInvalid</code>
	 * parameter is set to <code>Properties.ENSURE_PARENT_FOLDER_EXISTS</code>,
	 * the parent folder will be created and a reference to an inexistant file
	 * (in a existant folder) will be returned.
	 * <br />
	 * The return value is guaranteed non-<code>null</code> provided neither of
	 * the <code>defaultValue</code>s are <code>null</code>.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @param defaultIfMissing the file to use if the property is missing (use
	 *                         <code>Properties.FILE_PARAM_MUST_EXIST</code> to
	 *                         cause an exception in such a situation)
	 * @param defaultIfInvalid the file to use if the property does not specify
	 *                         a valid file (use
	 *                         <code>Properties.FILE_MUST_EXIST</code> to cause
	 *                         an exception in such a situation)
	 * @return file reference of the location specified by the given property
	 * @throws PropertyException the property does not exist or is invalid
	 */
	public File getFileProperty( String key, File defaultIfMissing, File defaultIfInvalid ) throws PropertyException
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
		{
			if (defaultIfMissing == FILE_PARAM_MUST_EXIST)
				throw new PropertyException( _sourceDescription, key );
			else
				return defaultIfMissing;
		}

		try
		{
			return Strings.parseFile( defaultIfInvalid, stringValue );
		}
		catch (ParseValueException e)
		{
			throw new PropertyException( _sourceDescription, key, stringValue, e );
		}
	}

	/**
	 * Parse a file reference from the property with given key, returning one
	 * of the default values if the property is missing or invalid.
	 * <br />
	 * If the property with given key does not exist the path specified by the
	 * <code>defaultIfMissing</code> parameter will be used in place of the
	 * property value.
	 * <br />
	 * If the value of the property with given key does not specify a valid
	 * file path and the <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.FILE_MUST_EXIST</code>, an exception will be thrown.
	 * <br />
	 * If the value of the property with given key does not specify a valid
	 * file path and the <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.ENSURE_FILE_EXISTS</code>, a new, empty file will be
	 * created and a reference to that new file returned.
	 * <br />
	 * If the value of the property with given key specifies a path to a file
	 * in a parent folder that does not exist and the
	 * <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.PARENT_FOLDER_MUST_EXIST</code>, an exception will be
	 * thrown.
	 * <br />
	 * If the value of the property with given key specifies a path to a file
	 * in a parent folder that does not exist and the
	 * <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.ENSURE_PARENT_FOLDER_EXISTS</code>, the parent folder
	 * will be created and a reference to an inexistant file (in an
	 * existant folder) will be returned.
	 * <br />
	 * The return value is guaranteed non-<code>null</code> provided neither of
	 * the <code>defaultValue</code>s are <code>null</code>. If
	 * <code>defaultIfInvalid</code> is set to
	 * <code>Properties.ENSURE_FILE_EXISTS</code> then the return value is
	 * also guaranteed to reference a file that exists.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @param defaultIfMissing the path of the file to use if the property is
	 *                         missing
	 * @param defaultIfInvalid the file to use if the property does not
	 *                         specify a valid file (use
	 *                         <code>Properties.FILE_MUST_EXIST</code> to
	 *                         cause an exception in such a situation)
	 * @return reference to the file specified by the given property
	 * @throws PropertyException the property does not exist or is invalid
	 */
	public File getFileProperty( String key, String defaultIfMissing, File defaultIfInvalid ) throws PropertyException
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
			stringValue = defaultIfMissing;

		try
		{
			return Strings.parseFile( defaultIfInvalid, stringValue );
		}
		catch (ParseValueException e)
		{
			throw new PropertyException( _sourceDescription, key, stringValue, e );
		}
	}

	/**
	 * Parse a file reference from the property with given key, returning the
	 * default value if the property is missing or invalid.
	 * <br />
	 * If the property with given key does not exist or does not specify a valid
	 * file then the default value will be returned.
	 * <br />
	 * The return value is guaranteed non-<code>null</code> provided the
	 * <code>defaultValue</code> is not <code>null</code>.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @param defaultValue the file to use if the property is missing or invalid
	 * @return file reference of the location specified by the given property
	 */
	public File getFileProperty( String key, File defaultValue )
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
			return defaultValue;

		return Strings.parseFile( stringValue, defaultValue );
	}

	/**
	 * Parse a file reference from the property with given key, using the
	 * default file path if the property is missing or invalid.
	 * <br />
	 * If the property with given key does not exist or does not specify a valid
	 * file then the default value will be returned.
	 * <br />
	 * The return value is guaranteed non-<code>null</code> provided the
	 * <code>defaultValue</code> is not <code>null</code>.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @param defaultPath the file path to use if the property is missing or
	 *                    invalid
	 * @return file reference of the location specified by the given property
	 */
	public File getFileProperty( String key, String defaultPath )
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
			stringValue = defaultPath;

		return Strings.parseFile( stringValue, ENSURE_FILE_EXISTS );
	}

	/**
	 * Parse a file reference from the property with given key.
	 * <br />
	 * The property with given key  should specify the path to a file, either
	 * absolute or relative to the <code>baseFolder</code>.
	 * <br />
	 * If the property with given key does not exist an exception will be
	 * thrown.
	 * <br />
	 * If the value of the property with given key does not specify a valid file
	 * path then an exception will be thrown.
	 * <br />
	 * The return value is guaranteed non-<code>null</code>.
	 * <br />
	 * <br />
	 * @param baseFolder the base/parent folder to use if the file path defined
	 *                   by the <code>string</code> is relative, rather than an
	 *                   absolute path
	 * @param key the name of the property to be read
	 * @return file reference of the location specified by the given property
	 * @throws PropertyException the property does not exist or is invalid
	 */
	public File getFileProperty( File baseFolder, String key ) throws PropertyException
	{
		return getFileProperty( baseFolder, key, FILE_PARAM_MUST_EXIST, FILE_MUST_EXIST );
	}

	/**
	 * Parse a file reference from the property with given key, returning one of
	 * the default values if the property is missing or invalid.
	 * <br />
	 * The property with given key should specify the path to a file, either
	 * absolute or relative to the <code>baseFolder</code>.
	 * <br />
	 * If the property with given key does not exist and the
	 * <code>defaultIfMissing</code> parameter is set to
	 * <code>Properties.FILE_PARAM_MUST_EXIST</code> then an exception will
	 * be thrown. Otherwise the <code>defaultIfMissing</code> value is returned.
	 * <br />
	 * If the value of the property with given key does not specify a valid file
	 * path and the <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.FILE_MUST_EXIST</code>, an exception will be thrown.
	 * <br />
	 * If the value of the property with given key does not specify a valid file
	 * path and the <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.ENSURE_FILE_EXISTS</code>, a new, empty file will be
	 * created and a reference to that new file returned.
	 * <br />
	 * If the value of the property with given key specifies a path to a file in
	 * a folder that does not exist and the <code>defaultIfInvalid</code>
	 * parameter is set to <code>Properties.PARENT_FOLDER_MUST_EXIST</code>, an
	 * exception will be thrown.
	 * <br />
	 * If the value of the property with given key specifies a path to a file in
	 * a folder that does not exist and the <code>defaultIfInvalid</code>
	 * parameter is set to <code>Properties.ENSURE_PARENT_FOLDER_EXISTS</code>,
	 * the parent folder will be created and a reference to an inexistant file
	 * (in a existant folder) will be returned.
	 * <br />
	 * The return value is guaranteed non-<code>null</code> provided neither of
	 * the <code>defaultValue</code>s are <code>null</code>.
	 * <br />
	 * <br />
	 * @param baseFolder the base/parent folder to use if the file path defined
	 *                   by the <code>string</code> is relative, rather than an
	 *                   absolute path
	 * @param key the name of the property to be read
	 * @param defaultIfMissing the file to use if the property is missing (use
	 *                         <code>Properties.FILE_PARAM_MUST_EXIST</code> to
	 *                         cause an exception in such a situation)
	 * @param defaultIfInvalid the file to use if the property does not specify
	 *                         a valid file (use
	 *                         <code>Properties.FILE_MUST_EXIST</code> to cause
	 *                         an exception in such a situation)
	 * @return file reference of the location specified by the given property
	 * @throws PropertyException the property does not exist or is invalid
	 */
	public File getFileProperty( File baseFolder, String key, File defaultIfMissing, File defaultIfInvalid ) throws PropertyException
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
		{
			if (defaultIfMissing == FILE_PARAM_MUST_EXIST)
				throw new PropertyException( _sourceDescription, key );
			else
				return defaultIfMissing;
		}

		try
		{
			return Strings.parseFile( defaultIfInvalid, stringValue, baseFolder );
		}
		catch (ParseValueException e)
		{
			throw new PropertyException( _sourceDescription, key, stringValue, e );
		}
	}

	/**
	 * Parse a file reference from the property with given key, returning one
	 * of the default values if the property is missing or invalid.
	 * <br />
	 * The property with given key should specify the path to a file, either
	 * absolute or relative to the <code>baseFolder</code>.
	 * <br />
	 * If the property with given key does not exist the path specified by the
	 * <code>defaultIfMissing</code> parameter will be used in place of the
	 * property value.
	 * <br />
	 * If the value of the property with given key does not specify a valid
	 * file path and the <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.FILE_MUST_EXIST</code>, an exception will be thrown.
	 * <br />
	 * If the value of the property with given key does not specify a valid
	 * file path and the <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.ENSURE_FILE_EXISTS</code>, a new, empty file will be
	 * created and a reference to that new file returned.
	 * <br />
	 * If the value of the property with given key specifies a path to a file
	 * in a parent folder that does not exist and the
	 * <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.PARENT_FOLDER_MUST_EXIST</code>, an exception will be
	 * thrown.
	 * <br />
	 * If the value of the property with given key specifies a path to a file
	 * in a parent folder that does not exist and the
	 * <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.ENSURE_PARENT_FOLDER_EXISTS</code>, the parent folder
	 * will be created and a reference to an inexistant file (in an
	 * existant folder) will be returned.
	 * <br />
	 * The return value is guaranteed non-<code>null</code> provided neither of
	 * the <code>defaultValue</code>s are <code>null</code>. If
	 * <code>defaultIfInvalid</code> is set to
	 * <code>Properties.ENSURE_FILE_EXISTS</code> then the return value is
	 * also guaranteed to reference a file that exists.
	 * <br />
	 * <br />
	 * @param baseFolder the base/parent folder to use if the file path defined
	 *                   by the <code>string</code> is relative, rather than an
	 *                   absolute path
	 * @param key the name of the property to be read
	 * @param defaultIfMissing the path of the file to use if the property is
	 *                         missing
	 * @param defaultIfInvalid the file to use if the property does not
	 *                         specify a valid file (use
	 *                         <code>Properties.FILE_MUST_EXIST</code> to
	 *                         cause an exception in such a situation)
	 * @return reference to the file specified by the given property
	 * @throws PropertyException the property does not exist or is invalid
	 */
	public File getFileProperty( File baseFolder, String key, String defaultIfMissing, File defaultIfInvalid ) throws PropertyException
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
			stringValue = defaultIfMissing;

		try
		{
			return Strings.parseFile( defaultIfInvalid, stringValue, baseFolder );
		}
		catch (ParseValueException e)
		{
			throw new PropertyException( _sourceDescription, key, stringValue, e );
		}
	}

	/**
	 * Parse a file reference from the property with given key, returning the
	 * default value if the property is missing or invalid.
	 * <br />
	 * The property with given key  should specify the path to a file, either
	 * absolute or relative to the <code>baseFolder</code>.
	 * <br />
	 * If the property with given key does not exist or does not specify a valid
	 * file then the default value will be returned.
	 * <br />
	 * The return value is guaranteed non-<code>null</code> provided the
	 * <code>defaultValue</code> is not <code>null</code>.
	 * <br />
	 * <br />
	 * @param baseFolder the base/parent folder to use if the file path defined
	 *                   by the <code>string</code> is relative, rather than an
	 *                   absolute path
	 * @param key the name of the property to be read
	 * @param defaultValue the file to use if the property is missing or invalid
	 * @return file reference of the location specified by the given property
	 */
	public File getFileProperty( File baseFolder, String key, File defaultValue )
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
			return defaultValue;

		return Strings.parseFile( stringValue, baseFolder, defaultValue );
	}

	/**
	 * Assign a file path to a given property.
	 * <br />
	 * <br />
	 * @param key the name of the property to be stored
	 * @param value the new value of the property
	 */
	public void setFileProperty( String key, File value )
	{
		setProperty( key, value.getPath() );
	}

	/**** ****    GET FOLDER PROPERTY    **** ****/

	/**
	 * Parse a folder reference from the property with given key.
	 * <br />
	 * If the property with given key does not exist an exception will be
	 * thrown.
	 * <br />
	 * If the value of the property with given key does not specify a valid
	 * folder  path then an exception will be thrown.
	 * <br />
	 * The return value is guaranteed non-<code>null</code>. However, no
	 * guarantee is made about the existance of the folder referenced by the
	 * <code>defaultValue</code>. Use
	 * <code>getFolderProperty( String key, String defaultIfMissing, File defaultIfInvalid )</code>
	 * and set <code>defaultIfInvalid</code> to
	 * <code>Properties.ENSURE_FOLDER_EXISTS</code> to be sure the return value
	 * references a folder that exists.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @return file reference of the location specified by the given property
	 * @throws PropertyException the property does not exist or is invalid
	 */
	public File getFolderProperty( String key ) throws PropertyException
	{
		return getFolderProperty( key, FOLDER_PARAM_MUST_EXIST, FOLDER_MUST_EXIST );
	}

	/**
	 * Parse a folder reference from the property with given key, returning one
	 * of the default values if the property is missing or invalid.
	 * <br />
	 * If the property with given key does not exist and the
	 * <code>defaultIfMissing</code> parameter is set to
	 * <code>Properties.FOLDER_PARAM_MUST_EXIST</code> then an exception will
	 * be thrown. Otherwise the <code>defaultIfMissing</code> value is returned.
	 * <br />
	 * If the value of the property with given key does not specify a valid
	 * folder path and the <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.FOLDER_MUST_EXIST</code>, an exception will be thrown.
	 * <br />
	 * If the value of the property with given key does not specify a valid
	 * folder path and the <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.ENSURE_FOLDER_EXISTS</code>, a new, empty folder will be
	 * created and a reference to that new folder returned.
	 * <br />
	 * If the value of the property with given key specifies a path to a folder
	 * in a parent folder that does not exist and the
	 * <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.PARENT_FOLDER_MUST_EXIST</code>, an exception will be
	 * thrown.
	 * <br />
	 * If the value of the property with given key specifies a path to a folder
	 * in a parent folder that does not exist and the
	 * <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.ENSURE_PARENT_FOLDER_EXISTS</code>, the parent folder
	 * will be created and a reference to an inexistant sub-folder (of an
	 * existant folder) will be returned.
	 * <br />
	 * The return value is guaranteed non-<code>null</code> provided neither of
	 * the <code>defaultValue</code>s are <code>null</code>. However, no
	 * guarantee is made about the existance of the folder referenced by the
	 * <code>defaultValue</code>. Use
	 * <code>getFolderProperty( String key, String defaultIfMissing, File defaultIfInvalid )</code>
	 * and set <code>defaultIfInvalid</code> to
	 * <code>Properties.ENSURE_FOLDER_EXISTS</code> to be sure the return value
	 * references a folder that exists.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @param defaultIfMissing the folder to use if the property is missing (use
	 *                         <code>Properties.FOLDER_PARAM_MUST_EXIST</code>
	 *                         to cause an exception in such a situation)
	 * @param defaultIfInvalid the folder to use if the property does not
	 *                         specify a valid folder (use
	 *                         <code>Properties.FOLDER_MUST_EXIST</code> to
	 *                         cause an exception in such a situation)
	 * @return file reference of the folder  specified by the given property
	 * @throws PropertyException the property does not exist or is invalid
	 */
	public File getFolderProperty( String key, File defaultIfMissing, File defaultIfInvalid ) throws PropertyException
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
		{
			if (defaultIfMissing == FOLDER_PARAM_MUST_EXIST)
				throw new PropertyException( _sourceDescription, key );
			else
				return defaultIfMissing;
		}

		try
		{
			return Strings.parseFolder( defaultIfInvalid, stringValue );
		}
		catch (ParseValueException e)
		{
			throw new PropertyException( _sourceDescription, key, stringValue, e );
		}
	}

	/**
	 * Parse a folder reference from the property with given key, returning one
	 * of the default values if the property is missing or invalid.
	 * <br />
	 * If the property with given key does not exist the path specified by the
	 * <code>defaultIfMissing</code> parameter will be used in place of the
	 * property value.
	 * <br />
	 * If the value of the property with given key does not specify a valid
	 * folder path and the <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.FOLDER_MUST_EXIST</code>, an exception will be thrown.
	 * <br />
	 * If the value of the property with given key does not specify a valid
	 * folder path and the <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.ENSURE_FOLDER_EXISTS</code>, a new, empty folder will be
	 * created and a reference to that new folder returned.
	 * <br />
	 * If the value of the property with given key specifies a path to a folder
	 * in a parent folder that does not exist and the
	 * <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.PARENT_FOLDER_MUST_EXIST</code>, an exception will be
	 * thrown.
	 * <br />
	 * If the value of the property with given key specifies a path to a folder
	 * in a parent folder that does not exist and the
	 * <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.ENSURE_PARENT_FOLDER_EXISTS</code>, the parent folder
	 * will be created and a reference to an inexistant sub-folder (of an
	 * existant folder) will be returned.
	 * <br />
	 * The return value is guaranteed non-<code>null</code> provided neither of
	 * the <code>defaultValue</code>s are <code>null</code>. If
	 * <code>defaultIfInvalid</code> is set to
	 * <code>Properties.ENSURE_FOLDER_EXISTS</code> then the return value is
	 * also guaranteed to reference a folder that exists.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @param defaultIfMissing the path of the folder to use if the property is
	 *                         missing
	 * @param defaultIfInvalid the folder to use if the property does not
	 *                         specify a valid folder (use
	 *                         <code>Properties.FOLDER_MUST_EXIST</code> to
	 *                         cause an exception in such a situation)
	 * @return file reference of the folder specified by the given property
	 * @throws PropertyException the property does not exist or is invalid
	 */
	public File getFolderProperty( String key, String defaultIfMissing, File defaultIfInvalid ) throws PropertyException
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
			stringValue = defaultIfMissing;

		try
		{
			return Strings.parseFolder( defaultIfInvalid, stringValue );
		}
		catch (ParseValueException e)
		{
			throw new PropertyException( _sourceDescription, key, stringValue, e );
		}
	}

	/**
	 * Parse a folder reference from the property with given key, returning the
	 * default value if the property is missing or invalid.
	 * <br />
	 * If the property with given key does not exist or does not specify a valid
	 * folder then the default value will be returned.
	 * <br />
	 * The return value is guaranteed non-<code>null</code> provided the
	 * <code>defaultValue</code> is not <code>null</code>. However, no guarantee
	 * is made about the existance of the folder referenced by the
	 * <code>defaultValue</code>. Use
	 * <code>getFolderProperty( String key, String defaultIfMissing, File defaultIfInvalid )</code>
	 * and set <code>defaultIfInvalid</code> to
	 * <code>Properties.ENSURE_FOLDER_EXISTS</code> to be sure the return value
	 * references a folder that exists.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @param defaultValue the folder to use if the property is missing or
	 *                     invalid
	 * @return file reference of the folder location specified by the given
	 *              property
	 */
	public File getFolderProperty( String key, File defaultValue )
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
			return defaultValue;

		return Strings.parseFolder( stringValue, defaultValue );
	}

	/**
	 * Parse a folder reference from the property with given key.
	 * <br />
	 * The property with given key should specify the path to a folder, either
	 * absolute or relative to the <code>baseFolder</code>.
	 * <br />
	 * If the property with given key does not exist an exception will be
	 * thrown.
	 * <br />
	 * If the value of the property with given key does not specify a valid
	 * folder path then an exception will be thrown.
	 * <br />
	 * The return value is guaranteed non-<code>null</code>. However, no
	 * guarantee is made about the existance of the folder referenced by the
	 * <code>defaultValue</code>. Use
	 * <code>getFolderProperty( File baseFolder, String key, String defaultIfMissing, File defaultIfInvalid )</code>
	 * and set <code>defaultIfInvalid</code> to
	 * <code>Properties.ENSURE_FOLDER_EXISTS</code> to be sure the return value
	 * references a folder that exists.
	 * <br />
	 * <br />
	 * @param baseFolder the base/parent folder to use if the folder path
	 *                   defined by the <code>string</code> is relative, rather
	 *                   than an absolute path
	 * @param key the name of the property to be read
	 * @return file reference of the folder location specified by the given
	 *              property
	 * @throws PropertyException the property does not exist or is invalid
	 */
	public File getFolderProperty( File baseFolder, String key ) throws PropertyException
	{
		return getFolderProperty( baseFolder, key, FOLDER_PARAM_MUST_EXIST, FOLDER_MUST_EXIST );
	}

	/**
	 * Parse a folder reference from the property with given key, returning one
	 * of the default values if the property is missing or invalid.
	 * <br />
	 * The property with given key should specify the path to a folder, either
	 * absolute or relative to the <code>baseFolder</code>.
	 * <br />
	 * If the property with given key does not exist and the
	 * <code>defaultIfMissing</code> parameter is set to
	 * <code>Properties.FOLDER_PARAM_MUST_EXIST</code> then an exception will
	 * be thrown. Otherwise the <code>defaultIfMissing</code> value is returned.
	 * <br />
	 * If the value of the property with given key does not specify a valid
	 * folder path and the <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.FOLDER_MUST_EXIST</code>, an exception will be thrown.
	 * <br />
	 * If the value of the property with given key does not specify a valid
	 * folder path and the <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.ENSURE_FOLDER_EXISTS</code>, a new, empty folder will be
	 * created and a reference to that new folder returned.
	 * <br />
	 * If the value of the property with given key specifies a path to a folder
	 * in a parent folder that does not exist and the
	 * <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.PARENT_FOLDER_MUST_EXIST</code>, an exception will be
	 * thrown.
	 * <br />
	 * If the value of the property with given key specifies a path to a folder
	 * in a parent folder that does not exist and the
	 * <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.ENSURE_PARENT_FOLDER_EXISTS</code>, the parent folder
	 * will be created and a reference to an inexistant sub-folder (of an
	 * existant folder) will be returned.
	 * <br />
	 * The return value is guaranteed non-<code>null</code> provided neither of
	 * the <code>defaultValue</code>s are <code>null</code>. However, no
	 * guarantee is made about the existance of the folder referenced by the
	 * <code>defaultValue</code>. Use
	 * <code>getFolderProperty( File baseFolder, String key, String defaultIfMissing, File defaultIfInvalid )</code>
	 * and set <code>defaultIfInvalid</code> to
	 * <code>Properties.ENSURE_FOLDER_EXISTS</code> to be sure the return value
	 * references a folder that exists.
	 * <br />
	 * <br />
	 * @param baseFolder the base/parent folder to use if the folder path
	 *                   defined by the <code>string</code> is relative, rather
	 *                   than an absolute path
	 * @param key the name of the property to be read
	 * @param defaultIfMissing the folder to use if the property is missing (use
	 *                         <code>Properties.FOLDER_PARAM_MUST_EXIST</code>
	 *                         to cause an exception in such a situation)
	 * @param defaultIfInvalid the folder to use if the property does not
	 *                         specify a valid folder (use
	 *                         <code>Properties.FOLDER_MUST_EXIST</code> to
	 *                         cause an exception in such a situation)
	 * @return file reference of the folder specified by the given property
	 * @throws PropertyException the property does not exist or is invalid
	 */
	public File getFolderProperty( File baseFolder, String key, File defaultIfMissing, File defaultIfInvalid ) throws PropertyException
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
		{
			if (defaultIfMissing == FOLDER_PARAM_MUST_EXIST)
				throw new PropertyException( _sourceDescription, key );
			else
				return defaultIfMissing;
		}

		try
		{
			return Strings.parseFolder( defaultIfInvalid, stringValue, baseFolder );
		}
		catch (ParseValueException e)
		{
			throw new PropertyException( _sourceDescription, key, stringValue, e );
		}
	}

	/**
	 * Parse a folder reference from the property with given key, returning one
	 * of the default values if the property is missing or invalid.
	 * <br />
	 * The property with given key should specify the path to a folder, either
	 * absolute or relative to the <code>baseFolder</code>.
	 * <br />
	 * If the property with given key does not exist, the path specified by the
	 * <code>defaultIfMissing</code> parameter will be used in place of the
	 * property value.
	 * <br />
	 * If the value of the property with given key does not specify a valid
	 * folder path and the <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.FOLDER_MUST_EXIST</code>, an exception will be thrown.
	 * <br />
	 * If the value of the property with given key does not specify a valid
	 * folder path and the <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.ENSURE_FOLDER_EXISTS</code>, a new, empty folder will be
	 * created and a reference to that new folder returned.
	 * <br />
	 * If the value of the property with given key specifies a path to a folder
	 * in a parent folder that does not exist and the
	 * <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.PARENT_FOLDER_MUST_EXIST</code>, an exception will be
	 * thrown.
	 * <br />
	 * If the value of the property with given key specifies a path to a folder
	 * in a parent folder that does not exist and the
	 * <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.ENSURE_PARENT_FOLDER_EXISTS</code>, the parent folder
	 * will be created and a reference to an inexistant sub-folder (of an
	 * existant folder) will be returned.
	 * <br />
	 * The return value is guaranteed non-<code>null</code> provided neither of
	 * the <code>defaultValue</code>s are <code>null</code>. If
	 * <code>defaultIfInvalid</code> is set to
	 * <code>Properties.ENSURE_FOLDER_EXISTS</code> then the return value is
	 * also guaranteed to reference a folder that exists.
	 * <br />
	 * <br />
	 * @param baseFolder the base/parent folder to use if the folder path
	 *                   defined by the <code>string</code> is relative, rather
	 *                   than an absolute path
	 * @param key the name of the property to be read
	 * @param defaultIfMissing the path of the folder to use if the property is
	 *                         missing
	 * @param defaultIfInvalid the folder to use if the property does not
	 *                         specify a valid folder (use
	 *                         <code>Properties.FOLDER_MUST_EXIST</code> to
	 *                         cause an exception in such a situation)
	 * @return file reference of the folder specified by the given property
	 * @throws PropertyException the property does not exist or is invalid
	 */
	public File getFolderProperty( File baseFolder, String key, String defaultIfMissing, File defaultIfInvalid ) throws PropertyException
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
			stringValue = defaultIfMissing;

		try
		{
			return Strings.parseFolder( defaultIfInvalid, stringValue, baseFolder );
		}
		catch (ParseValueException e)
		{
			throw new PropertyException( _sourceDescription, key, stringValue, e );
		}
	}

	/**
	 * Parse a folder reference from the property with given key, returning the
	 * default value if the property is missing or invalid.
	 * <br />
	 * The property with given key should specify the path to a folder, either
	 * absolute or relative to the <code>baseFolder</code>.
	 * <br />
	 * If the property with given key does not exist or does not specify a valid
	 * folder then the default value will be returned.
	 * <br />
	 * The return value is guaranteed non-<code>null</code> provided the
	 * <code>defaultValue</code> is not <code>null</code>. However, no guarantee
	 * is made about the existance of the folder referenced by the
	 * <code>defaultValue</code>. Use
	 * <code>getFolderProperty( File baseFolder, String key, String defaultIfMissing, File defaultIfInvalid )</code>
	 * and set <code>defaultIfInvalid</code> to
	 * <code>Properties.ENSURE_FOLDER_EXISTS</code> to be sure the return value
	 * references a folder that exists.
	 * <br />
	 * <br />
	 * @param baseFolder the base/parent folder to use if the folder path
	 *                   defined by the <code>string</code> is relative, rather
	 *                   than an absolute path
	 * @param key the name of the property to be read
	 * @param defaultValue the folder to use if the property is missing or
	 *                     invalid
	 * @return file reference of the folder location specified by the given
	 *              property
	 */
	public File getFolderProperty( File baseFolder, String key, File defaultValue )
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
			return defaultValue;

		return Strings.parseFolder( stringValue, baseFolder, defaultValue );
	}

	/**
	 * Assign a folder path to a given property.
	 * <br />
	 * <br />
	 * @param key the name of the property to be stored
	 * @param value the new value of the property
	 */
	public void setFolderProperty( String key, File value )
	{
		setProperty( key, value.getPath() );
	}

	/**** ****    GET CLASS PROPERTY    **** ****/

	/**
	 * Parse a Java class from the property with given key.
	 * <br />
	 * If the property with given key does not exist an exception will be
	 * thrown.
	 * <br />
	 * If the value of the property with given key does not specify a valid
	 * class path then an exception will be thrown.
	 * <br />
	 * The return value is guaranteed non-<code>null</code>.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @return the Java class specified by the given property
	 * @throws PropertyException the property does not exist or is invalid
	 */
	public Class getClassProperty( String key ) throws PropertyException
	{
		return getClassProperty( key, CLASS_PARAM_MUST_EXIST, CLASS_MUST_BE_VALID );
	}

	/**
	 * Parse a Java class from the property with given key, returning one
	 * of the default values if the property is missing or invalid.
	 * <br />
	 * If the property with given key does not exist and the
	 * <code>defaultIfMissing</code> parameter is set to
	 * <code>Properties.CLASS_PARAM_MUST_EXIST</code> then an exception will
	 * be thrown. Otherwise the <code>defaultIfMissing</code> value is returned.
	 * <br />
	 * If the value of the property with given key does not specify a valid
	 * class path and the <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.CLASS_MUST_EXIST</code>, an exception will be thrown.
	 * <br />
	 * The return value is guaranteed non-<code>null</code> provided neither of
	 * the <code>defaultValue</code>s are <code>null</code>.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @param defaultIfMissing the class to use if the property is missing (use
	 *                         <code>Properties.CLASS_PARAM_MUST_EXIST</code>
	 *                         to cause an exception in such a situation)
	 * @param defaultIfInvalid the class to use if the property does not
	 *                         specify a valid Java class (use
	 *                         <code>Properties.CLASS_MUST_BE_VALID</code> to
	 *                         cause an exception in such a situation)
	 * @return the Java class specified by the given property
	 * @throws PropertyException the property does not exist or is invalid
	 */
	public Class getClassProperty( String key, Class defaultIfMissing, Class defaultIfInvalid ) throws PropertyException
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
		{
			if (defaultIfMissing == CLASS_PARAM_MUST_EXIST)
				throw new PropertyException( _sourceDescription, key );
			else
				return defaultIfMissing;
		}

		try
		{
			return Strings.parseClass( defaultIfInvalid, stringValue );
		}
		catch (ParseValueException e)
		{
			throw new PropertyException( _sourceDescription, key, stringValue, e );
		}
	}

	/**
	 * Parse a Java class from the property with given key, returning the
	 * default value if the property is missing or invalid.
	 * <br />
	 * If the property with given key does not exist or does not specify a valid
	 * folder then the default value will be returned.
	 * <br />
	 * The return value is guaranteed non-<code>null</code> provided the
	 * <code>defaultValue</code> is not <code>null</code>.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @param defaultValue the class to use if the property is missing or
	 *                     invalid
	 * @return the Java class specified by the given property
	 */
	public Class getClassProperty( String key, Class defaultValue )
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
			return defaultValue;

		return Strings.parseClass( stringValue, defaultValue );
	}

	/**
	 * Parse a Java class from the property with given key.
	 * <br />
	 * The property with given key should specify the path of a class, either
	 * absolute or relative to the <code>basePackage</code>.
	 * <br />
	 * If the property with given key does not exist an exception will be
	 * thrown.
	 * <br />
	 * If the value of the property with given key does not specify a valid
	 * folder path then an exception will be thrown.
	 * <br />
	 * The return value is guaranteed non-<code>null</code>.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @param basePackage the base package path to use if the class name defined
	 *                    by the parameter is not an absolute path
	 * @return the Java class specified by the given property
	 * @throws PropertyException the property does not exist or is invalid
	 */
	public Class getClassProperty( String key, String basePackage ) throws PropertyException
	{
		return getClassProperty( key, basePackage, CLASS_PARAM_MUST_EXIST, CLASS_MUST_BE_VALID );
	}

	/**
	 * Parse a Java class from the property with given key, returning one of
	 * the default values if the property is missing or invalid.
	 * <br />
	 * The property with given key should specify the path of a class, either
	 * absolute or relative to the <code>basePackage</code>.
	 * <br />
	 * If the property with given key does not exist and the
	 * <code>defaultIfMissing</code> parameter is set to
	 * <code>Properties.CLASS_PARAM_MUST_EXIST</code> then an exception will
	 * be thrown. Otherwise the <code>defaultIfMissing</code> value is returned.
	 * <br />
	 * If the value of the property with given key does not specify a valid
	 * class path and the <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.CLASS_MUST_EXIST</code>, an exception will be thrown.
	 * <br />
	 * The return value is guaranteed non-<code>null</code> provided neither of
	 * the <code>defaultValue</code>s are <code>null</code>.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @param basePackage the base package path to use if the class name defined
	 *                    by the parameter is not an absolute path
	 * @param defaultIfMissing the class to use if the property is missing (use
	 *                         <code>Properties.CLASS_PARAM_MUST_EXIST</code>
	 *                         to cause an exception in such a situation)
	 * @param defaultIfInvalid the class to use if the property does not
	 *                         specify a valid Java class (use
	 *                         <code>Properties.CLASS_MUST_BE_VALID</code> to
	 *                         cause an exception in such a situation)
	 * @return the Java class specified by the given property
	 * @throws PropertyException the property does not exist or is invalid
	 */
	public Class getClassProperty( String key, String basePackage, Class defaultIfMissing, Class defaultIfInvalid ) throws PropertyException
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
		{
			if (defaultIfMissing == CLASS_PARAM_MUST_EXIST)
				throw new PropertyException( _sourceDescription, key );
			else
				return defaultIfMissing;
		}

		try
		{
			return Strings.parseClass( defaultIfInvalid, stringValue, basePackage );
		}
		catch (ParseValueException e)
		{
			throw new PropertyException( _sourceDescription, key, stringValue, e );
		}
	}

	/**
	 * Parse a Java class from the property with given key, returning the
	 * default value if the property is missing or invalid.
	 * <br />
	 * The property with given key should specify the path of a class, either
	 * absolute or relative to the <code>basePackage</code>.
	 * <br />
	 * If the property with given key does not exist or does not specify a valid
	 * folder then the default value will be returned.
	 * <br />
	 * The return value is guaranteed non-<code>null</code> provided the
	 * <code>defaultValue</code> is not <code>null</code>.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @param basePackage the base package path to use if the class name defined
	 *                    by the parameter is not an absolute path
	 * @param defaultValue the folder to use if the property is missing or
	 *                     invalid
	 * @return the Java class specified by the given property
	 */
	public Class getClassProperty( String key, String basePackage, Class defaultValue )
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
			return defaultValue;

		return Strings.parseClass( stringValue, basePackage, defaultValue );
	}

	/**
	 * Assign the name of a Java class to a given property.
	 * <br />
	 * <br />
	 * @param key the name of the property to be stored
	 * @param value the new value of the property
	 */
	public void setClassProperty( String key, Class value )
	{
		setProperty( key, value.getName() );
	}

	/**** ****    GET ENUMERATION PROPERTY    **** ****/

	/**
	 * Parse an enumeration value from the property with given key.
	 * <br />
	 * If the property with given key does not exist an exception will be
	 * thrown.
	 * <br />
	 * If the value of the property with given key does not equate to any of the
	 * values in the given <code>enumeration</code> then an exception will be
	 * thrown.
	 * <br />
	 * The return value is guaranteed non-<code>null</code>.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @param enumeration the enumeration type
	 * @return the enumeration value, taken from the given enumeration,
	 *         identified by the value of the given property
	 * @throws PropertyException the property does not exist or is invalid
	 */
	public EnumValue getEnumProperty( String key, Enumeration enumeration ) throws PropertyException
	{
		return getEnumProperty( key, enumeration, ENUM_PARAM_MUST_EXIST, ENUM_VALUE_MUST_BE_VALID );
	}

	/**
	 * Parse an enumeration value from the property with given key, returning
	 * one of the default values if the property is missing or invalid.
	 * <br />
	 * If the property with given key does not exist and the
	 * <code>defaultIfMissing</code> parameter is set to
	 * <code>Properties.ENUM_PARAM_MUST_EXIST</code> then an exception will
	 * be thrown. Otherwise the <code>defaultIfMissing</code> value is returned.
	 * <br />
	 * If the value of the property with given key does not equate to any of the
	 * values in the given <code>enumeration</code> and the
	 * <code>defaultIfInvalid</code> parameter is set to
	 * <code>Properties.ENUM_VALUE_MUST_BE_VALID</code>, an exception will be
	 * thrown.
	 * <br />
	 * The return value is guaranteed non-<code>null</code> provided neither of
	 * the <code>defaultValue</code>s are <code>null</code>.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @param enumeration the enumeration type
	 * @param defaultIfMissing the enumeration value to use if the property is
	 *                         missing (use
	 *                         <code>Properties.ENUM_PARAM_MUST_EXIST</code>
	 *                         to cause an exception in such a situation)
	 * @param defaultIfInvalid the enumeration value to use if the property does
	 *                         not equate to one of the enumeration values (use
	 *                         <code>Properties.ENUM_VALUE_MUST_BE_VALID</code>
	 *                         to cause an exception in such a situation)
	 * @return the enumeration value, taken from the given enumeration,
	 *         identified by the value of the given property
	 * @throws PropertyException the property does not exist or is invalid
	 */
	public EnumValue getEnumProperty( String key, Enumeration enumeration, EnumValue defaultIfMissing, EnumValue defaultIfInvalid ) throws PropertyException
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
		{
			if (defaultIfMissing == ENUM_PARAM_MUST_EXIST)
				throw new PropertyException( _sourceDescription, key );
			else
				return defaultIfMissing;
		}

		try
		{
			return Strings.parseEnumValue( defaultIfInvalid, stringValue, enumeration );
		}
		catch (ParseValueException e)
		{
			throw new PropertyException( _sourceDescription, key, stringValue, e );
		}
	}

	/**
	 * Parse an enumeration value from the property with given key, returning
	 * the default value if the property is missing or invalid.
	 * <br />
	 * If the value of the property with given key does not equate to any of the
	 * values in the given <code>enumeration</code> then the default value will
	 * be returned.
	 * <br />
	 * The return value is guaranteed non-<code>null</code> provided the
	 * <code>defaultValue</code> is not <code>null</code>.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @param enumeration the enumeration type
	 * @param defaultValue the enumeration value to use if the property is
	 *                     missing or invalid
	 * @return the enumeration value, taken from the given enumeration,
	 *         identified by the value of the given property
	 */
	public EnumValue getEnumProperty( String key, Enumeration enumeration, EnumValue defaultValue )
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
			return defaultValue;

		return Strings.parseEnumValue( stringValue, enumeration, defaultValue );
	}
	
	/**
	 * Assign an enumeration value to a given property.
	 * <br />
	 * <br />
	 * @param key the name of the property to be stored
	 * @param value the new value of the property
	 */
	public void setEnumProperty( String key, EnumValue value )
	{
		setProperty( key, value.toString() );
	}

	/**** ****    GET COLOR PROPERTY    **** ****/

	/**
	 * Parse a <code>Color</code> from the property with given key.
	 * <br />
	 * <br />
	 * The property value should specify a colour in either:
	 * <br />
	 * &#149; decimal format:<code> "[AAA] RRR GGG BBB" </code>where
	 * <code>AAA</code>, <code>RRR</code>, <code>GGG</code> and <code>BBB</code>
	 * are 16-bit decimal values (0-255) for the alpha, red, green and blue
	 * components of the colour (alpha optional) each separated by one or more
	 * whitespace characters
	 * <br />
	 * &#149; hexadecimal format:<code> "#[AA]RRGGBB" </code> where
	 * <code>AA</code>, <code>RR</code>, <code>GG</code> and <code>BB</code> are
	 * 16-bit hex values for the alpha, red, green and blue components of the
	 * colour (alpha optional)
	 * <br />
	 * &#149; equivalent decimal value: a single decimal value that equates to
	 * a hexadecimal value in <code> "#[AA]RRGGBB" </code> format
	 * <br />
	 * <br />
	 * If the property with given key does not exist and the
	 * <code>defaultIfMissing</code> parameter is set to
	 * <code>Properties.COLOR_PARAM_MUST_EXIST</code> then an exception will
	 * be thrown. Otherwise the <code>defaultIfMissing</code> value is returned.
	 * <br />
	 * If the value of the property with given key does not conform to any of
	 * the above formats and the <code>defaultIfInvalid</code> parameter is set
	 * to <code>Properties.COLOR_VALUE_MUST_BE_VALID</code>, an exception will
	 * be thrown.
	 * <br />
	 * <br />
	 * If the property with given key does not exist an exception will be
	 * thrown.
	 * <br />
	 * If the value of the property with given key does not equate to any of the
	 * values in the given <code>enumeration</code> then an exception will be
	 * thrown.
	 * <br />
	 * The return value is guaranteed non-<code>null</code>.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @return a <code>Color</code> object for the color specified by the
	 *         property value
	 * @throws PropertyException the property does not exist or is invalid
	 */
	public Color getColorProperty( String key ) throws PropertyException
	{
		return getColorProperty( key, COLOR_PARAM_MUST_EXIST, COLOR_MUST_BE_VALID );
	}

	/**
	 * Parse a <code>Color</code> from the property with given key, returning
	 * one of the default values if the property is missing or invalid.
	 * <br />
	 * <br />
	 * The property value should specify a colour in either:
	 * <br />
	 * &#149; decimal format:<code> "[AAA] RRR GGG BBB" </code>where
	 * <code>AAA</code>, <code>RRR</code>, <code>GGG</code> and <code>BBB</code>
	 * are 16-bit decimal values (0-255) for the alpha, red, green and blue
	 * components of the colour (alpha optional) each separated by one or more
	 * whitespace characters
	 * <br />
	 * &#149; hexadecimal format:<code> "#[AA]RRGGBB" </code> where
	 * <code>AA</code>, <code>RR</code>, <code>GG</code> and <code>BB</code> are
	 * 16-bit hex values for the alpha, red, green and blue components of the
	 * colour (alpha optional)
	 * <br />
	 * &#149; equivalent decimal value: a single decimal value that equates to
	 * a hexadecimal value in <code> "#[AA]RRGGBB" </code> format
	 * <br />
	 * <br />
	 * If the property with given key does not exist and the
	 * <code>defaultIfMissing</code> parameter is set to
	 * <code>Properties.COLOR_PARAM_MUST_EXIST</code> then an exception will
	 * be thrown. Otherwise the <code>defaultIfMissing</code> value is returned.
	 * <br />
	 * If the value of the property with given key does not conform to any of
	 * the above formats and the <code>defaultIfInvalid</code> parameter is set
	 * to <code>Properties.COLOR_VALUE_MUST_BE_VALID</code>, an exception will
	 * be thrown.
	 * <br />
	 * The return value is guaranteed non-<code>null</code> provided neither of
	 * the <code>defaultValue</code>s are <code>null</code>.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @param defaultIfMissing the value to use if the property is missing (use
	 *                         <code>Properties.ENUM_PARAM_MUST_EXIST</code>
	 *                         to cause an exception in such a situation)
	 * @param defaultIfInvalid the value to use if the property does not equate
	 *                         to one of the enumeration values (use
	 *                         <code>Properties.ENUM_VALUE_MUST_BE_VALID</code>
	 *                         to cause an exception in such a situation)
	 * @return a <code>Color</code> object for the color specified by the
	 *         property value, or the default value if the property value is
	 *         invalid
	 * @throws PropertyException the property does not exist or is invalid
	 */
	public Color getColorProperty( String key, Color defaultIfMissing, Color defaultIfInvalid ) throws PropertyException
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
		{
			if (defaultIfMissing == COLOR_PARAM_MUST_EXIST)
				throw new PropertyException( _sourceDescription, key );
			else
				return defaultIfMissing;
		}

		try
		{
			return Strings.parseColor( defaultIfInvalid, stringValue );
		}
		catch (ParseValueException e)
		{
			throw new PropertyException( _sourceDescription, key, stringValue, e );
		}
	}

	/**
	 * Parse a <code>Color</code> from the property with given key, returning
	 * the default value if the property is missing or invalid.
	 * <br />
	 * <br />
	 * The property value should specify a colour in either:
	 * <br />
	 * &#149; decimal format:<code> "[AAA] RRR GGG BBB" </code>where
	 * <code>AAA</code>, <code>RRR</code>, <code>GGG</code> and <code>BBB</code>
	 * are 16-bit decimal values (0-255) for the alpha, red, green and blue
	 * components of the colour (alpha optional) each separated by one or more
	 * whitespace characters
	 * <br />
	 * &#149; hexadecimal format:<code> "#[AA]RRGGBB" </code> where
	 * <code>AA</code>, <code>RR</code>, <code>GG</code> and <code>BB</code> are
	 * 16-bit hex values for the alpha, red, green and blue components of the
	 * colour (alpha optional)
	 * <br />
	 * &#149; equivalent decimal value: a single decimal value that equates to
	 * a hexadecimal value in <code> "#[AA]RRGGBB" </code> format
	 * <br />
	 * <br />
	 * If the value of the property with given key does not conform to any of
	 * the above formats then the default value will be returned.
	 * <br />
	 * The return value is guaranteed non-<code>null</code> provided the
	 * <code>defaultValue</code> is not <code>null</code>.
	 * <br />
	 * <br />
	 * @param key the name of the property to be read
	 * @param defaultValue the value to use if the property is missing or
	 *                     invalid
	 * @return a <code>Color</code> object for the color specified by the
	 *         property value, or the default value if the property value is
	 *         invalid
	 */
	public Color getColorProperty( String key, Color defaultValue )
	{
		String stringValue = getProperty( key );
		if (stringValue == null)
			return defaultValue;
		
		return Strings.parseColor( stringValue, defaultValue );
	}
	
	/**
	 * Assign a color value to a given property.
	 * <br />
	 * <br />
	 * @param key the name of the property to be stored
	 * @param value the new value of the property
	 */
	public void setColorProperty( String key, Color value )
	{
		setProperty( key, Strings.serializeColor( value ) );
	}
	
	
	/*
	 * Inner classes
	 */
	
	/**
	 * Helper class required for exported constant CLASS_PARAM_MUST_EXIST, which
	 * must be unique/different to avoid confusion with any other value a client
	 * may use.
	 */
	private static class ClassParamMustExist
	{
	}
	
	/**
	 * Helper class required for exported constant ENUM_PARAM_MUST_EXIST,
	 * which must be unique/different to avoid confusion with any other value a
	 * client may use.
	 */
	private static class EnumParamMustExist extends EnumValue
	{
		/** Class serialisation version id. */
		private static final long serialVersionUID = 1L;
		
		/** A list of enumeration values. */
		public static final Enumeration ENUMERATION = new Enumeration();
		
		/*
		 * Overloaded methods of EnumValue base class.
		 */
		
		public Enumeration getEnumeration()
		{
			return ENUMERATION;
		}
		
		/*
		 * Constructor
		 */
		
		/**
		 * The default constructor, protected to prevent use outside this class.
		 */
		protected EnumParamMustExist()
		{
		}
	}
	
}


/*
 * Change History
 *
 * $Log: Properties.java,v $
 * Revision 1.6  2006/10/15 21:20:10  Steve
 * Added some usage instructions in the class JavaDoc.
 * Fixed the name of exception thrown in the JavaDoc of some getter methods.
 *
 * Revision 1.5  2006/09/17 08:34:46  Steve
 * Fixed method name for getEnumProperty.
 *
 * Revision 1.4  2006/09/06 20:52:02  Steve
 * Added legal disclaimer.
 *
 * Revision 1.3  2006/09/02 17:23:33  Steve
 * Written JavaDoc for getColor methods.
 * Fixed JavaDoc errors.
 *
 * Revision 1.2  2006/08/15 21:42:13  Steve
 * Added methods to set non-string values.
 * Added getter and setter methods for Color values.
 * The properties table can be now be associated with a File - default load & store methods synchronise data between the file and the table in memory.
 * Added JavaDoc comments for exported constants.
 *
 * Revision 1.1  2006/06/04 18:54:46  Steve
 * Initial revision.
 *
 */
