/*
 * IsolatedKeyEventProcessor.java
 * Part of the B'am library - a collection of general-purpose utility classes.
 * 
 * Copyright (C)2006 Stephen Battey
 * Contact: resource@lucidviews.net
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

package net.lucidviews.util.gui.event;

import java.awt.event.KeyListener;
import java.awt.event.KeyEvent;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;


/**
 * Listens for key events and isolates a single key press.
 * <br />
 * For each key event a thread is started which causes a delay before sending
 * the actual notification for that key press. If another key event is received
 * within the tolerance (delay) time slot then both events are cancelled.
 * <br />
 * <br />
 * @author Stephen Battey
 * @version $Revision: 1.3 $
 * @since 1.0
 */
public class IsolatedKeyEventProcessor implements KeyListener
{
	
	/*
	 * Instance variables
	 */
	
	protected Collection      _listeners;
	protected long            _tolerance;
	protected IsolationThread _isolationThread;
	
	
	/*
	 * Constructor(s)
	 */
	
	/**
	 * Create a processor for key events which only accepts isolated key events
	 * (those which occur apart by least the tolerance).
	 * <br />
	 * <br />
	 * @param tolerance the isolation tolerance, in milli-seconds
	 */
	public IsolatedKeyEventProcessor( long tolerance )
	{
		_listeners = new ArrayList();
		_tolerance = tolerance;
		_isolationThread = null;
	}
	
	
	/*
	 * Implementation of KeyListener interface
	 */
	
	public void keyTyped( KeyEvent e )
	{
	}
	
	public void keyPressed( KeyEvent e )
	{
		startIsolationThread( e );
	}
	
	public void keyReleased( KeyEvent e )
	{
	}
	
	
	/*
	 * Implementation of IsolatedKeyEventListener notifier
	 */
	
	/**
	 * Register a listener that wants to be notified of isolated key press events.
	 * <br />
	 * <br />
	 * @param listener object that is to receive notifications
	 */
	public void addIsolatedKeyEventListener( IsolatedKeyEventListener listener )
	{
		synchronized (_listeners)
		{
			if (!_listeners.contains( listener ))
				_listeners.add( listener );
		}
	}

	/**
	 * Un-register a listener that does not want to be notified of isolated key
	 * press events.
	 * <br />
	 * <br />
	 * @param listener object that is no longer to receive notifications
	 */
	public void removeIsolatedKeyEventListener( IsolatedKeyEventListener listener )
	{
		synchronized (_listeners)
		{
			if (_listeners.contains( listener ))
				_listeners.remove( listener );
		}
	}
	
	/**
	 * Notify all listeners that an isolated key press has occurred.
	 * <br />
	 * <br />
	 * @param keyEvent the key event associated with the isolated key press
	 */
	protected void notifyIsolatedKeyPress( KeyEvent keyEvent )
	{
		/* notify all listeners of the isolated key press */
		synchronized (_listeners)
		{
			Iterator itr = _listeners.iterator();
			while (itr.hasNext())
				((IsolatedKeyEventListener)itr.next()).isolatedKeyPress( keyEvent );
		}
	}
	
	
	/*
	 * Methods
	 */
	
	/**
	 * Start a new isolation thread to verify a given key event is isolated.
	 * <br />
	 * This method is called from the <code>keyPressed</code> method. It will
	 * pass control of processing the key event to the
	 * <code>IsolationThread::run</code> method.
	 * <br />
	 * <br />
	 * @param keyEvent the incoming key event
	 */
	protected synchronized void startIsolationThread( KeyEvent keyEvent )
	{
		if (_isolationThread == null)
		/* start a new isolation thread */
		{
			_isolationThread = new IsolationThread( this, keyEvent, _tolerance );
			_isolationThread.start();
		}
		else
		/* check if the key events are for the same key */
		{
			if (keyEvent.getKeyCode() == _isolationThread._keyEvent.getKeyCode())
			{
				/* accept current thread as still isolated, but don't start a new thread
				   for the repeat key */
			}
			else
			{
				/* neither the current or new threads are isolated */
				_isolationThread._isIsolated = false;
				
				_isolationThread = new IsolationThread( this, keyEvent, _tolerance );
				_isolationThread._isIsolated = false;
				_isolationThread.start();
			}
		}
	}
	
	/**
	 * Stop an isolation thread.
	 * <br />
	 * This method is called from the <code>IsolationThread::run</code> method. It
	 * will check the event has been successfully isolated before calling the
	 * <code>notifyIsolatedKeyPress</code> which will pass control of processing
	 * the key event to any listeners.
	 * <br />
	 * <br />
	 * @param isolationThread the thread which may have isolated a key event
	 */
	protected synchronized void stopIsolationThread( IsolationThread isolationThread )
	{
		/* check if the key event was isolated */
		if (isolationThread._isIsolated)
			notifyIsolatedKeyPress( isolationThread._keyEvent );

		/* if this is the current thread, remove reference to the current thread */
		if (_isolationThread == isolationThread)
			_isolationThread = null;
	}
	
	
	/*
	 * Helper class
	 */
	
	public static class IsolationThread extends Thread
	{
		
		/*
		 * Instance variables
		 */
		
		protected IsolatedKeyEventProcessor _isolationProcessor;
		protected KeyEvent                  _keyEvent;
		protected long                      _delay;
		protected boolean                   _isIsolated;
		
		
		/*
		 * Constructor(s)
		 */
		
		public IsolationThread( IsolatedKeyEventProcessor isolationProcessor,
		                        KeyEvent keyEvent, long delay )
		{
			_isolationProcessor = isolationProcessor;
			_keyEvent = keyEvent;
			_delay = delay;
			_isIsolated = true;
		}
		
		
		/*
		 * Overridden methods of Thread base class
		 */
		
		public void run()
		{
			/* wait the desired tolerance */
			try
			{
				Thread.sleep( _delay );
			}
			catch (InterruptedException e)
			{
			}
			
			_isolationProcessor.stopIsolationThread( this );
		}
		
	}
	
}


/*
 * Change History
 *
 * $Log: IsolatedKeyEventProcessor.java,v $
 * Revision 1.3  2006/09/06 21:07:54  Steve
 * Added legal disclaimer.
 *
 * Revision 1.2  2006/06/07 20:34:29  Steve
 * Brought into Bam library:
 *  - moved into lucidviews package
 *  - added CVS headers/footers
 *
 */
