/*
 * JImage.java
 * Part of the B'am library - a collection of general-purpose utility classes.
 * 
 * Copyright (C)2005, 2006 Stephen Battey
 * Contact: resource@lucidviews.net
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

package net.lucidviews.util.gui.swing;

import net.lucidviews.util.image.DimensionImageObserver;
import net.lucidviews.util.image.RepaintImageObserver;

import javax.swing.border.Border;
import javax.swing.JComponent;

import java.awt.image.ImageObserver;
import java.awt.Image;
import java.awt.Point;
import java.awt.Dimension;
import java.awt.Insets;
import java.awt.Graphics;
import java.awt.Toolkit;
import java.net.URL;
import java.io.File;


/**
 * A UI component that displays an image.
 * <br />
 * To display an image in a GUI, simply create an new instance of
 * <code>JImage</code> specifying an <code>Image</code> or the location of a
 * file containing an image:
 * <pre>
 * Image logoImage = Toolkit.getDefaultToolkit.createImage( "./icons/logo.gif" );
 *
 * JImage logoPane = new JImage( logoImage, new Dimension( 80, 50 ) );
 * JImage localPicture = new JImage( "C:/MyPictures/Photo.jpg", JImage.MAINTAIN_SIZE );
 * JImage remotePicture = new JImage( new URL( "http://www.mysite.com/Photo.jpg" ), JImage.SHRINK_TO_AREA_MAINTAIN_ASPECT );
 *
 * JPanel contentPane = ...;
 * contentPane.add( logoPane );
 * contentPane.add( localPicture );
 * contentPane.add( remotePicture );</pre>
 * <br />
 * The functionality required to load a file from a local file or remote URL is
 * encapsulated in this class. The image data will load in the background and
 * the UI component will be refreshed once the image has fully loaded.
 * <br />
 * <br />
 * You can change the image displayed at any time by calling one of the
 * <code>setImage</code> methods.
 * <br />
 * If you intend to change the image frequently (e.g. switch between two images)
 * it is recommended you load those images in your client module to ensure the
 * <code>Image</code> objects persist in memory. Using the
 * <code>setImage( <i>filename</i> )</code> method will cause each image to load
 * on every call to <code>setImage</code>. Holding the <code>Image</code> object
 * in memory and calling <code>setImage( Image )</code> will ensure the image
 * is loaded just once and provide a smoother display.
 * <br />
 * <br />
 * @see java.awt.Image
 * @see java.awt.Toolkit#createImage
 * @author Stephen Battey
 * @version $Revision: 1.5 $
 * @since 1.0
 */
public class JImage extends JComponent implements DimensionImageObserver.Listener
{
	
	/*
	 * Exported constants
	 */
	
	/* Special value used to indicate that the image should maintain its
	 * original size regardless of the area available in which to display it. */
	public static final Dimension MAINTAIN_SIZE = new Dimension();
	
	/* Special value used to indicate that the image should be stretched or
	 * shrunk to fit the available area. */
	public static final Dimension FIT_TO_AREA = new Dimension();
	
	/* Special value used to indicate that the image should be stretched or
	 * shrunk to fit the available area but should maintain its aspect ratio. */
	public static final Dimension FIT_TO_AREA_MAINTAIN_ASPECT = new Dimension();
	
	/* Special value used to indicate that the image should be shrunk to fit the
	 * available area but should maintain its aspect ratio. */
	public static final Dimension SHRINK_TO_AREA_MAINTAIN_ASPECT = new Dimension();
	
	
	/*
	 * Local constants
	 */
	
	/** A list of file extensions supported by the Java Image renderer. */
	protected static final String[] SUPPORTED_IMAGE_EXTNS = { ".jpg", ".gif", ".jpeg" };
	
	/** Special value used to indicate the image size has not been set. */
	protected static final Dimension UNDEFINED_SIZE = new Dimension();
	
	/** An <code>Insets</code> object that defines no border. */
	protected static final Insets ZERO_INSET = new Insets( 0, 0, 0, 0 );
	
	
	/*
	 * Instance variables
	 */
	
	/** The image to be displayed. */
	protected Image _image = null;
	
	/** The (original) size of the image. */
	protected Dimension _imageSize = null;
	
	/** The algorithm used to determine the size of the displayed image based
	 *  upon the available area in which to display it. */
	protected Dimension _displaySetting = null;
	
	/** The size of the rendered image. */
	protected Dimension _displaySize = UNDEFINED_SIZE;
	
	/** The algorithm used to scale the original image to the displayed
	 *  (re-sized) image. The algorthim can be changed to improve quality or
	 *  performance. */
	protected int _displayScaling = Image.SCALE_DEFAULT;
	
	/** Horizontal alignment of the image within the available space
	 *  (0.0: left; 1.0: right). */
	protected float _xAlignment = LEFT_ALIGNMENT;
	
	/** Vertical alignment of the image within the available space
	 *  (0.0: top; 1.0: bottom) */
	protected float _yAlignment = TOP_ALIGNMENT;
	
	/** The image that gets displayed
	 *  (a re-sized copy of <code>_image</code>) */
	protected Image _displayImage = null;
	
	/** The pixel position of the top-right corner of the image relative to the
	 *  top-right corner of the available area. */
	protected Point _displayPosition = new Point();
	
	/** Flag set when the paint method is called and reset once the image has
	 *  been successfully painted (i.e. flag is on if the image needs
	 *  repainting). */
	protected boolean _paintRequested = false;
	
	
	/*
	 * Constructors
	 */
	
	/**
	 * Create a component that displays an image, but leave the image undefined.
	 * <br />
	 * The component will remain blank until the image has been set with a call
	 * to the <code>setImage</code> method.
	 */
	public JImage()
	{
	}
	
	/**
	 * Create a component that displays an image.
	 * <br />
	 * <br />
	 * @param image the image to display
	 * @param size either an absolute value or one of the constants exported by
	 *             <code>JImage</code>
	 * @see #MAINTAIN_SIZE
	 * @see #FIT_TO_AREA
	 * @see #FIT_TO_AREA_MAINTAIN_ASPECT
	 * @see #SHRINK_TO_AREA_MAINTAIN_ASPECT
	 */
	public JImage( Image image, Dimension size )
	{
		setImage( image, size, false );
	}
	
	/**
	 * Create a component that displays an image taken from a given file.
	 * <br />
	 * <br />
	 * @param filename the name of the file that stores the image; the file
	 *                 must be in one of the supported formats (JPG or GIF)
	 * @param size either an absolute value or one of the constants exported by
	 *             <code>JImage</code>
	 * @see #MAINTAIN_SIZE
	 * @see #FIT_TO_AREA
	 * @see #FIT_TO_AREA_MAINTAIN_ASPECT
	 * @see #SHRINK_TO_AREA_MAINTAIN_ASPECT
	 */
	public JImage( String filename, Dimension size )
	{
		this( Toolkit.getDefaultToolkit().createImage( filename ), size );
	}
	
	/**
	 * Create a GUI component that displays an image, using the given scaling
	 * algorithm, whenever necessary, to adjust the size of the image to fit
	 * the available space.
	 * <br />
	 * <br />
	 * @param image the image to display
	 * @param size either an absolute value or one of the constants exported by
	 *             <code>JImage</code>
	 * @param scalingAlgorithm flags to indicate the type of algorithm to use
     *                         for image re-scaling
	 * @see #MAINTAIN_SIZE
	 * @see #FIT_TO_AREA
	 * @see #FIT_TO_AREA_MAINTAIN_ASPECT
	 * @see #SHRINK_TO_AREA_MAINTAIN_ASPECT
     * @see java.awt.Image#SCALE_DEFAULT
     * @see java.awt.Image#SCALE_FAST
     * @see java.awt.Image#SCALE_SMOOTH
     * @see java.awt.Image#SCALE_REPLICATE
     * @see java.awt.Image#SCALE_AREA_AVERAGING
	 */
	public JImage( Image image, Dimension size, int scalingAlgorithm )
	{
		setImage( image, size, scalingAlgorithm, false );
	}
	
	/**
	 * Create a GUI component that displays an image taken from a given file,
	 * using the given scaling algorithm, whenever necessary, to adjust the size
	 * of the image to fit the available space.
	 * <br />
	 * <br />
	 * @param filename the name of the file that stores the image; the file
	 *                 must be in one of the supported formats (JPG or GIF)
	 * @param size either an absolute value or one of the constants exported by
	 *             <code>JImage</code>
	 * @param scalingAlgorithm flags to indicate the type of algorithm to use
     *                         for image re-scaling
	 * @see #MAINTAIN_SIZE
	 * @see #FIT_TO_AREA
	 * @see #FIT_TO_AREA_MAINTAIN_ASPECT
	 * @see #SHRINK_TO_AREA_MAINTAIN_ASPECT
     * @see java.awt.Image#SCALE_DEFAULT
     * @see java.awt.Image#SCALE_FAST 
     * @see java.awt.Image#SCALE_SMOOTH
     * @see java.awt.Image#SCALE_REPLICATE
     * @see java.awt.Image#SCALE_AREA_AVERAGING
	 */
	public JImage( String filename, Dimension size, int scalingAlgorithm )
	{
		this( Toolkit.getDefaultToolkit().createImage( filename ), size, scalingAlgorithm );
	}
	
	
	/*
	 * Methods
	 */
	
	/**
	 * Set the alignment of the image within the available space.
	 * <br />
	 * The alignment will affect the position of the image after it has been
	 * scaled to fit the available area. If any excess space remains (due to the
	 * difference in ratio aspect between the image and available area) then the
	 * alignment setting will cause the image to be aligned to one of the edges
	 * or in the centre. 
	 * <br /> 
	 * <br />
	 * @param xAlignment horizontal alignment (0.0: left; 1.0: right)
	 * @param yAlignment vertical alignment (0.0: top; 1.0: bottom)
	 */
	public void setImageAlignment( float xAlignment, float yAlignment )
	{
		_xAlignment = xAlignment;
		_yAlignment = yAlignment;
	}
	
	/**
	 * Get the <code>Image</code> object that represents the image being
	 * displayed in this component.
	 * <br /> 
	 * <br />
	 * @return the image being displayed or <code>null</code> if the image has
	 *         not been set
	 */
	public Image getImage()
	{
		return _image;
	}
	
	/**
	 * Set the image to be displayed in this component.
	 * <br />
	 * This component will be repainted as soon as the image has loaded to
	 * ensure the new image is rendered.
	 * <br />
	 * The image will be rendered at the <code>size</code> specified or, if
	 * <code>size</code> is set to one of the constants exported by this class,
	 * then the image will be re-sized taking into account the space available
	 * for the image.
	 * <br />
	 * <br />
	 * @param image the image to be displayed
	 * @param size either an absolute value or one of the constants exported by
	 *             <code>JImage</code>
	 * @see #MAINTAIN_SIZE
	 * @see #FIT_TO_AREA
	 * @see #FIT_TO_AREA_MAINTAIN_ASPECT
	 * @see #SHRINK_TO_AREA_MAINTAIN_ASPECT
	 */
	public void setImage( Image image, Dimension size )
	{
		setImage( image, size, Image.SCALE_DEFAULT, true );
	}
	
	/**
	 * Set the image to be displayed in this component.
	 * <br />
	 * This component will be repainted as soon as the image has loaded to
	 * ensure the new image is rendered.
	 * <br />
	 * The image will be rendered at the <code>size</code> specified or, if
	 * <code>size</code> is set to one of the constants exported by this class,
	 * then the image will be re-sized taking into account the space available
	 * for the image.
	 * <br />
	 * <br />
	 * @param filename the name of the file that stores the image; the file
	 *                 must be in one of the supported formats (JPG or GIF)
	 * @param size either an absolute value or one of the constants exported by
	 *             <code>JImage</code>
	 * @see #MAINTAIN_SIZE
	 * @see #FIT_TO_AREA
	 * @see #FIT_TO_AREA_MAINTAIN_ASPECT
	 * @see #SHRINK_TO_AREA_MAINTAIN_ASPECT
	 */
	public void setImage( String filename, Dimension size )
	{
		setImage( Toolkit.getDefaultToolkit().createImage( filename ), size, Image.SCALE_DEFAULT, true );
	}
	
	/**
	 * Set the image to be displayed in this component.
	 * <br />
	 * This component will be repainted as soon as the image has loaded to
	 * ensure the new image is rendered.
	 * <br />
	 * The image will be rendered at the <code>size</code> specified or, if
	 * <code>size</code> is set to one of the constants exported by this class,
	 * then the image will be re-sized taking into account the space available
	 * for the image.
	 * <br />
	 * <br />
	 * @param imageURL the URL of the file that stores the image; the file
	 *                 must be in one of the supported formats (JPG or GIF)
	 * @param size either an absolute value or one of the constants exported by
	 *             <code>JImage</code>
	 * @see #MAINTAIN_SIZE
	 * @see #FIT_TO_AREA
	 * @see #FIT_TO_AREA_MAINTAIN_ASPECT
	 * @see #SHRINK_TO_AREA_MAINTAIN_ASPECT
	 */
	public void setImage( URL imageURL, Dimension size )
	{
		setImage( Toolkit.getDefaultToolkit().createImage( imageURL ), size, Image.SCALE_DEFAULT, true );
	}
	
	/**
	 * Set the image to be displayed in this component.
	 * <br />
	 * This component will be repainted to ensure the new image is rendered.
	 * <br />
	 * The image will be rendered at the <code>size</code> specified or, if
	 * <code>size</code> is set to one of the constants exported by this class,
	 * then the image will be re-sized taking into account the space available
	 * for the image.
	 * <br />
	 * <br />
	 * @param imageData the binary data of the image to be displayed
	 * @param size either an absolute value or one of the constants exported by
	 *             <code>JImage</code>
	 * @see #MAINTAIN_SIZE
	 * @see #FIT_TO_AREA
	 * @see #FIT_TO_AREA_MAINTAIN_ASPECT
	 * @see #SHRINK_TO_AREA_MAINTAIN_ASPECT
	 */
	public void setImage( byte[] imageData, Dimension size )
	{
		setImage( Toolkit.getDefaultToolkit().createImage( imageData ), size, Image.SCALE_DEFAULT, true );
	}
	
	/**
	 * Set the image to be displayed in this component.
	 * <br />
	 * The image will be rendered at the <code>size</code> specified or, if
	 * <code>size</code> is set to one of the constants exported by this class,
	 * then the image will be re-sized taking into account the space available
	 * for the image.
	 * <br />
	 * <br />
	 * @param image the image to be displayed
	 * @param size either an absolute value or one of the constants exported by
	 *             <code>JImage</code>
	 * @param repaintNow indicates if the component should be refreshed to
	 *                   ensure the new image is rendered
	 * @see #MAINTAIN_SIZE
	 * @see #FIT_TO_AREA
	 * @see #FIT_TO_AREA_MAINTAIN_ASPECT
	 * @see #SHRINK_TO_AREA_MAINTAIN_ASPECT
	 */
	public synchronized void setImage( Image image, Dimension size, boolean repaintNow )
	{
		setImage( image, size, Image.SCALE_DEFAULT, repaintNow );
	}
	
	/**
	 * Set the image to be displayed in this component.
	 * <br />
	 * The image will be rendered at the <code>size</code> specified or, if
	 * <code>size</code> is set to one of the constants exported by this class,
	 * then the image will be re-sized taking into account the space available
	 * for the image.
	 * <br />
	 * <br />
	 * @param filename the name of the file that stores the image; the file
	 *                 must be in one of the supported formats (JPG or GIF)
	 * @param size either an absolute value or one of the constants exported by
	 *             <code>JImage</code>
	 * @param repaintNow indicates if the component should be refreshed to
	 *                   ensure the new image is rendered
	 * @see #MAINTAIN_SIZE
	 * @see #FIT_TO_AREA
	 * @see #FIT_TO_AREA_MAINTAIN_ASPECT
	 * @see #SHRINK_TO_AREA_MAINTAIN_ASPECT
	 */
	public void setImage( String filename, Dimension size, boolean repaintNow )
	{
		setImage( Toolkit.getDefaultToolkit().createImage( filename ), size, Image.SCALE_DEFAULT, repaintNow );
	}
	
	/**
	 * Set the image to be displayed in this component.
	 * <br />
	 * The image will be rendered at the <code>size</code> specified or, if
	 * <code>size</code> is set to one of the constants exported by this class,
	 * then the image will be re-sized taking into account the space available
	 * for the image.
	 * <br />
	 * <br />
	 * @param imageURL the URL of the file that stores the image; the file
	 *                 must be in one of the supported formats (JPG or GIF)
	 * @param size either an absolute value or one of the constants exported by
	 *             <code>JImage</code>
	 * @param repaintNow indicates if the component should be refreshed to
	 *                   ensure the new image is rendered
	 * @see #MAINTAIN_SIZE
	 * @see #FIT_TO_AREA
	 * @see #FIT_TO_AREA_MAINTAIN_ASPECT
	 * @see #SHRINK_TO_AREA_MAINTAIN_ASPECT
	 */
	public void setImage( URL imageURL, Dimension size, boolean repaintNow )
	{
		setImage( Toolkit.getDefaultToolkit().createImage( imageURL ), size, Image.SCALE_DEFAULT, repaintNow );
	}
	
	/**
	 * Set the image to be displayed in this component.
	 * <br />
	 * The image will be rendered at the <code>size</code> specified or, if
	 * <code>size</code> is set to one of the constants exported by this class,
	 * then the image will be re-sized taking into account the space available
	 * for the image.
	 * <br />
	 * <br />
	 * @param imageData the binary data of the image to be displayed
	 * @param size either an absolute value or one of the constants exported by
	 *             <code>JImage</code>
	 * @param repaintNow indicates if the component should be refreshed to
	 *                   ensure the new image is rendered
	 * @see #MAINTAIN_SIZE
	 * @see #FIT_TO_AREA
	 * @see #FIT_TO_AREA_MAINTAIN_ASPECT
	 * @see #SHRINK_TO_AREA_MAINTAIN_ASPECT
	 */
	public void setImage( byte[] imageData, Dimension size, boolean repaintNow )
	{
		setImage( Toolkit.getDefaultToolkit().createImage( imageData ), size, Image.SCALE_DEFAULT, repaintNow );
	}
	
	/**
	 * Set the image to be displayed in this component.
	 * <br />
	 * This component will be repainted as soon as the image has loaded to
	 * ensure the new image is rendered.
	 * <br />
	 * The image will be rendered at the <code>size</code> specified or, if
	 * <code>size</code> is set to one of the constants exported by this class,
	 * then the image will be re-sized taking into account the space available
	 * for the image.
	 * <br />
	 * If the image needs to be re-sized to fit into the available space (or
	 * to a specific, absolute size) then the <code>scalingAlgorithm</code>
	 * (<i>suggested</i> by the bitmap of constants exported by the
	 * <code>java.awt.Image</code> class) will be used.
	 * <br />
	 * <br />
	 * @param image the image to be displayed
	 * @param size either an absolute value or one of the constants exported by
	 *             <code>JImage</code>
	 * @param scalingAlgorithm flags to indicate the type of algorithm to use
     *                         for image re-scaling
	 * @see #MAINTAIN_SIZE
	 * @see #FIT_TO_AREA
	 * @see #FIT_TO_AREA_MAINTAIN_ASPECT
	 * @see #SHRINK_TO_AREA_MAINTAIN_ASPECT
     * @see java.awt.Image#SCALE_DEFAULT
     * @see java.awt.Image#SCALE_FAST 
     * @see java.awt.Image#SCALE_SMOOTH
     * @see java.awt.Image#SCALE_REPLICATE
     * @see java.awt.Image#SCALE_AREA_AVERAGING
	 */
	public void setImage( Image image, Dimension size, int scalingAlgorithm )
	{
		setImage( image, size, scalingAlgorithm, true );
	}
	
	/**
	 * Set the image to be displayed in this component.
	 * <br />
	 * This component will be repainted as soon as the image has loaded to
	 * ensure the new image is rendered.
	 * <br />
	 * The image will be rendered at the <code>size</code> specified or, if
	 * <code>size</code> is set to one of the constants exported by this class,
	 * then the image will be re-sized taking into account the space available
	 * for the image.
	 * <br />
	 * If the image needs to be re-sized to fit into the available space (or
	 * to a specific, absolute size) then the <code>scalingAlgorithm</code>
	 * (<i>suggested</i> by the bitmap of constants exported by the
	 * <code>java.awt.Image</code> class) will be used.
	 * <br />
	 * <br />
	 * @param filename the name of the file that stores the image; the file
	 *                 must be in one of the supported formats (JPG or GIF)
	 * @param size either an absolute value or one of the constants exported by
	 *             <code>JImage</code>
	 * @param scalingAlgorithm flags to indicate the type of algorithm to use
     *                         for image re-scaling
	 * @see #MAINTAIN_SIZE
	 * @see #FIT_TO_AREA
	 * @see #FIT_TO_AREA_MAINTAIN_ASPECT
	 * @see #SHRINK_TO_AREA_MAINTAIN_ASPECT
     * @see java.awt.Image#SCALE_DEFAULT
     * @see java.awt.Image#SCALE_FAST 
     * @see java.awt.Image#SCALE_SMOOTH
     * @see java.awt.Image#SCALE_REPLICATE
     * @see java.awt.Image#SCALE_AREA_AVERAGING
	 */
	public void setImage( String filename, Dimension size, int scalingAlgorithm )
	{
		setImage( Toolkit.getDefaultToolkit().createImage( filename ), size, scalingAlgorithm, true );
	}
	
	/**
	 * Set the image to be displayed in this component.
	 * <br />
	 * This component will be repainted as soon as the image has loaded to
	 * ensure the new image is rendered.
	 * <br />
	 * The image will be rendered at the <code>size</code> specified or, if
	 * <code>size</code> is set to one of the constants exported by this class,
	 * then the image will be re-sized taking into account the space available
	 * for the image.
	 * <br />
	 * If the image needs to be re-sized to fit into the available space (or
	 * to a specific, absolute size) then the <code>scalingAlgorithm</code>
	 * (<i>suggested</i> by the bitmap of constants exported by the
	 * <code>java.awt.Image</code> class) will be used.
	 * <br />
	 * <br />
	 * @param imageURL the URL of the file that stores the image; the file
	 *                 must be in one of the supported formats (JPG or GIF)
	 * @param size either an absolute value or one of the constants exported by
	 *             <code>JImage</code>
	 * @param scalingAlgorithm flags to indicate the type of algorithm to use
     *                         for image re-scaling
	 * @see #MAINTAIN_SIZE
	 * @see #FIT_TO_AREA
	 * @see #FIT_TO_AREA_MAINTAIN_ASPECT
	 * @see #SHRINK_TO_AREA_MAINTAIN_ASPECT
     * @see java.awt.Image#SCALE_DEFAULT
     * @see java.awt.Image#SCALE_FAST 
     * @see java.awt.Image#SCALE_SMOOTH
     * @see java.awt.Image#SCALE_REPLICATE
     * @see java.awt.Image#SCALE_AREA_AVERAGING
	 */
	public void setImage( URL imageURL, Dimension size, int scalingAlgorithm )
	{
 	  setImage( Toolkit.getDefaultToolkit().createImage( imageURL ), size, scalingAlgorithm, true );
	}
	
	/**
	 * Set the image to be displayed in this component.
	 * <br />
	 * This component will be repainted to ensure the new image is rendered.
	 * <br />
	 * The image will be rendered at the <code>size</code> specified or, if
	 * <code>size</code> is set to one of the constants exported by this class,
	 * then the image will be re-sized taking into account the space available
	 * for the image.
	 * <br />
	 * If the image needs to be re-sized to fit into the available space (or
	 * to a specific, absolute size) then the <code>scalingAlgorithm</code>
	 * (<i>suggested</i> by the bitmap of constants exported by the
	 * <code>java.awt.Image</code> class) will be used.
	 * <br />
	 * <br />
	 * @param imageData the binary data of the image to be displayed
	 * @param size either an absolute value or one of the constants exported by
	 *             <code>JImage</code>
	 * @param scalingAlgorithm flags to indicate the type of algorithm to use
     *                         for image re-scaling
	 * @see #MAINTAIN_SIZE
	 * @see #FIT_TO_AREA
	 * @see #FIT_TO_AREA_MAINTAIN_ASPECT
	 * @see #SHRINK_TO_AREA_MAINTAIN_ASPECT
     * @see java.awt.Image#SCALE_DEFAULT
     * @see java.awt.Image#SCALE_FAST 
     * @see java.awt.Image#SCALE_SMOOTH
     * @see java.awt.Image#SCALE_REPLICATE
     * @see java.awt.Image#SCALE_AREA_AVERAGING
 	 */
	public void setImage( byte[] imageData, Dimension size, int scalingAlgorithm )
	{
		setImage( Toolkit.getDefaultToolkit().createImage( imageData ), size, scalingAlgorithm, true );
	}
	
	/**
	 * Set the image to be displayed in this component.
	 * <br />
	 * The image will be rendered at the <code>size</code> specified or, if
	 * <code>size</code> is set to one of the constants exported by this class,
	 * then the image will be re-sized taking into account the space available
	 * for the image.
	 * <br />
	 * If the image needs to be re-sized to fit into the available space (or
	 * to a specific, absolute size) then the <code>scalingAlgorithm</code>
	 * (<i>suggested</i> by the bitmap of constants exported by the
	 * <code>java.awt.Image</code> class) will be used.
	 * <br />
	 * <br />
	 * @param image the image to be displayed
	 * @param size either an absolute value or one of the constants exported by
	 *             <code>JImage</code>
	 * @param scalingAlgorithm flags to indicate the type of algorithm to use
     *                         for image re-scaling
	 * @param repaintNow indicates if the component should be refreshed to
	 *                   ensure the new image is rendered
	 * @see #MAINTAIN_SIZE
	 * @see #FIT_TO_AREA
	 * @see #FIT_TO_AREA_MAINTAIN_ASPECT
	 * @see #SHRINK_TO_AREA_MAINTAIN_ASPECT
     * @see java.awt.Image#SCALE_DEFAULT
     * @see java.awt.Image#SCALE_FAST 
     * @see java.awt.Image#SCALE_SMOOTH
     * @see java.awt.Image#SCALE_REPLICATE
     * @see java.awt.Image#SCALE_AREA_AVERAGING
	 */
	public synchronized void setImage( Image image, Dimension size, int scalingAlgorithm, boolean repaintNow )
	{
		_image = image;
		_displaySetting = size;
		_displaySize = UNDEFINED_SIZE;
		_displayScaling = scalingAlgorithm;
		_displayImage = null;
		
		
		/* determine the size of the image (may not have loaded yet) */
		ImageObserver imgObserver = new DimensionImageObserver( this, _image );
		int imgWidth = _image.getWidth( imgObserver );
		int imgHeight = _image.getHeight( imgObserver );
		
		if (   (imgWidth != -1)
		    && (imgHeight != -1)  )
		{
			/* size of image is known; set component size now */
			setImageDimension( imgWidth, imgHeight );
		}
		/* otherwise the size of the image is unknown;
		   when the image has loaded 'imageUpdate' will be called and the size will then be set */
		
		
		if (repaintNow)
			/* paint the new image */
			repaint();
	}
	
	/**
	 * Set the image to be displayed in this component.
	 * <br />
	 * The image will be rendered at the <code>size</code> specified or, if
	 * <code>size</code> is set to one of the constants exported by this class,
	 * then the image will be re-sized taking into account the space available
	 * for the image.
	 * <br />
	 * If the image needs to be re-sized to fit into the available space (or
	 * to a specific, absolute size) then the <code>scalingAlgorithm</code>
	 * (<i>suggested</i> by the bitmap of constants exported by the
	 * <code>java.awt.Image</code> class) will be used.
	 * <br />
	 * <br />
	 * @param filename the name of the file that stores the image; the file
	 *                 must be in one of the supported formats (JPG or GIF)
	 * @param size either an absolute value or one of the constants exported by
	 *             <code>JImage</code>
	 * @param scalingAlgorithm flags to indicate the type of algorithm to use
     *                         for image re-scaling
	 * @param repaintNow indicates if the component should be refreshed to
	 *                   ensure the new image is rendered
	 * @see #MAINTAIN_SIZE
	 * @see #FIT_TO_AREA
	 * @see #FIT_TO_AREA_MAINTAIN_ASPECT
	 * @see #SHRINK_TO_AREA_MAINTAIN_ASPECT
     * @see java.awt.Image#SCALE_DEFAULT
     * @see java.awt.Image#SCALE_FAST 
     * @see java.awt.Image#SCALE_SMOOTH
     * @see java.awt.Image#SCALE_REPLICATE
     * @see java.awt.Image#SCALE_AREA_AVERAGING
 	 */
	public void setImage( String filename, Dimension size, int scalingAlgorithm, boolean repaintNow )
	{
		setImage( Toolkit.getDefaultToolkit().createImage( filename ), size, scalingAlgorithm, repaintNow );
	}
	
	/**
	 * Set the image to be displayed in this component.
	 * <br />
	 * The image will be rendered at the <code>size</code> specified or, if
	 * <code>size</code> is set to one of the constants exported by this class,
	 * then the image will be re-sized taking into account the space available
	 * for the image.
	 * <br />
	 * If the image needs to be re-sized to fit into the available space (or
	 * to a specific, absolute size) then the <code>scalingAlgorithm</code>
	 * (<i>suggested</i> by the bitmap of constants exported by the
	 * <code>java.awt.Image</code> class) will be used.
	 * <br />
	 * <br />
	 * @param imageURL the URL of the file that stores the image; the file
	 *                 must be in one of the supported formats (JPG or GIF)
	 * @param size either an absolute value or one of the constants exported by
	 *             <code>JImage</code>
	 * @param scalingAlgorithm flags to indicate the type of algorithm to use
     *                         for image re-scaling
	 * @param repaintNow indicates if the component should be refreshed to
	 *                   ensure the new image is rendered
	 * @see #MAINTAIN_SIZE
	 * @see #FIT_TO_AREA
	 * @see #FIT_TO_AREA_MAINTAIN_ASPECT
	 * @see #SHRINK_TO_AREA_MAINTAIN_ASPECT
     * @see java.awt.Image#SCALE_DEFAULT
     * @see java.awt.Image#SCALE_FAST 
     * @see java.awt.Image#SCALE_SMOOTH
     * @see java.awt.Image#SCALE_REPLICATE
     * @see java.awt.Image#SCALE_AREA_AVERAGING
	 */
	public void setImage( URL imageURL, Dimension size, int scalingAlgorithm, boolean repaintNow )
	{
		setImage( Toolkit.getDefaultToolkit().createImage( imageURL ), size, scalingAlgorithm, repaintNow );
	}
	
	/**
	 * Set the image to be displayed in this component.
	 * <br />
	 * The image will be rendered at the <code>size</code> specified or, if
	 * <code>size</code> is set to one of the constants exported by this class,
	 * then the image will be re-sized taking into account the space available
	 * for the image.
	 * <br />
	 * If the image needs to be re-sized to fit into the available space (or
	 * to a specific, absolute size) then the <code>scalingAlgorithm</code>
	 * (<i>suggested</i> by the bitmap of constants exported by the
	 * <code>java.awt.Image</code> class) will be used.
	 * <br />
	 * <br />
	 * @param imageData the binary data of the image to be displayed
	 * @param size either an absolute value or one of the constants exported by
	 *             <code>JImage</code>
	 * @param scalingAlgorithm flags to indicate the type of algorithm to use
     *                         for image re-scaling
	 * @param repaintNow indicates if the component should be refreshed to
	 *                   ensure the new image is rendered
	 * @see #MAINTAIN_SIZE
	 * @see #FIT_TO_AREA
	 * @see #FIT_TO_AREA_MAINTAIN_ASPECT
	 * @see #SHRINK_TO_AREA_MAINTAIN_ASPECT
     * @see java.awt.Image#SCALE_DEFAULT
     * @see java.awt.Image#SCALE_FAST 
     * @see java.awt.Image#SCALE_SMOOTH
     * @see java.awt.Image#SCALE_REPLICATE
     * @see java.awt.Image#SCALE_AREA_AVERAGING
	 */
	public void setImage( byte[] imageData, Dimension size, int scalingAlgorithm, boolean repaintNow )
	{
		setImage( Toolkit.getDefaultToolkit().createImage( imageData ), size, scalingAlgorithm, repaintNow );
	}
	
	
	/*
	 * Overridden methods of JPanel
	 */
	
	public void setBounds( int x, int y, int width, int height )
	{
		super.setBounds( x, y, width, height );
		
		if (_imageSize != null)
		{
			if (   (_displaySetting == FIT_TO_AREA)
			    || (_displaySetting == FIT_TO_AREA_MAINTAIN_ASPECT)
			    || (_displaySetting == SHRINK_TO_AREA_MAINTAIN_ASPECT)  )
			{
				/* resize image based on current size of panel */
				setDisplaySize( computeDisplaySize( _imageSize.width, _imageSize.height ) );
				
				/* repaint the image now it has been resized */
				repaint();
			}
		}
	}
	
	
	/*
	 * Local methods
	 */
	
	/**
	 * Compute the dimension of the scaled image.
	 * <br />
	 * This method will use the size of the original image and the size of the
	 * available area to compute the size of the displayed image. The original
	 * dimension will be stretched or squeezed to fit the available space
	 * according to the display setting.
	 * <br /> 
	 * <br />
	 * @param imageWidth the width of the original image
	 * @param imageHeight the height of the original image
	 * @return the dimension of the image when it is displayed
	 * @see #_displaySetting
	 */
	protected Dimension computeDisplaySize( int imageWidth, int imageHeight )
	{
		Dimension displaySize;
		
		if (_displaySetting == FIT_TO_AREA)
			displaySize = getSize();
		else
		{
			Insets insets = getBorderInsets();
			int areaWidth = getWidth() - insets.left - insets.right;
			int areaHeight = getHeight() - insets.top - insets.bottom;
			
			if (   (_displaySetting == SHRINK_TO_AREA_MAINTAIN_ASPECT)
			    && (imageWidth <= areaWidth)
			    && (imageHeight <= areaHeight)  )
			{
				/* image fits in display area, keep actual size */
				displaySize = new Dimension( imageWidth, imageHeight );
			}
			else
			{
				/* convert image size and available area to floating point numbers */
				float imageWidthF = (float)imageWidth;
				float imageHeightF = (float)imageHeight;
				float areaWidthF = (float)areaWidth;
				float areaHeightF = (float)areaHeight;
				
				/* compute scale factors in both x & y dimensions */
				float widthFactor = areaWidthF / imageWidthF;
				float heightFactor = areaHeightF / imageHeightF;
				
				/* scale image down to fit in most restricted plane */
				float scale = (widthFactor > heightFactor) ? heightFactor : widthFactor;
				displaySize = new Dimension( (int)(imageWidthF * scale), (int)(imageHeightF * scale) );
			}
		}
		
		return displaySize;
	}
	
	/**
	 * Set the size of the original image.
	 * <br />
	 * This method is called by the image observer that is monitoring the
	 * loading of the image object. This ensures the size is correct, as defined
	 * from the image file.
	 * <br /> 
	 * <br />
	 * @param imgWidth the width of the image to be displayed
	 * @param imgHeight the height of the image to be displayed
	 */
	public void setImageDimension( int imgWidth, int imgHeight )
	{
		_imageSize = new Dimension( imgWidth, imgHeight );
		
		if (_displaySetting == MAINTAIN_SIZE)
		{
			/* display image to actual size of image */
			setDisplaySize( new Dimension( imgWidth, imgHeight ) );
		}
		else if (   (_displaySetting == FIT_TO_AREA)
		         || (_displaySetting == FIT_TO_AREA_MAINTAIN_ASPECT)
		         || (_displaySetting == SHRINK_TO_AREA_MAINTAIN_ASPECT)  )
		{
			/* resize image based on current size of panel */
			setDisplaySize( computeDisplaySize( imgWidth, imgHeight ) );
		}
		else
		{
			/* display image to size requested */
			setDisplaySize( _displaySetting );
		}
	}
	
	/**
	 * Called to set the dimension of the image that is displayed in the
	 * component.
	 * <br />
	 * Unlike the dimension passed to the <code>setImage</code> method, this
	 * size must be an absolute dimension value (<b>NOT</b> one of the special
	 * constants exported by this class).
	 * <br />
	 * <br />
	 * @param size the size of the image when it is rendered
	 */
	protected void setDisplaySize( Dimension size )
	{
		/* min value is 1 pixel in either dimension */
		if (size.width < 1)
			size.width = 1;
		if (size.height < 1)
			size.height = 1;
		
		if (   (_displaySize != UNDEFINED_SIZE)
		    && (_displaySize.equals( size ))  )
			/* size has not changed, no need to re-scale and re-paint */
			return;
		
		/* remember the size of display */
		_displaySize = size;
		if (size.equals( _imageSize ))
			_displayImage = _image;
		else
			_displayImage = _image.getScaledInstance( size.width, size.height, _displayScaling );
		
		/* compute position of image given its size */
		Insets insets = getBorderInsets();
		if (_xAlignment == LEFT_ALIGNMENT)
			_displayPosition.x = insets.left;
		else
		{
			int xSpace = getWidth() - size.width - insets.left - insets.right;
			if (_xAlignment == CENTER_ALIGNMENT)
				_displayPosition.x = insets.left + (xSpace / 2);
			else
				_displayPosition.x = insets.left + xSpace;
		}
		if (_yAlignment == TOP_ALIGNMENT)
			_displayPosition.y = insets.top;
		else
		{
			int ySpace = getHeight() - size.height - insets.top - insets.bottom;
			if (_yAlignment == CENTER_ALIGNMENT)
				_displayPosition.y = insets.top + (ySpace / 2);
			else
				_displayPosition.y = insets.top + ySpace;
		}
		
		/* we would prefer to display the whole image */
		setPreferredSize( size );
		
		/* re-position components within the GUI to take account of the change of size */
		revalidate();
	}

	/**
	 * Returns an <code>Insets</code> model that describes the location of the
	 * inner edge of the border around this image component.
	 * <br />
	 * <br />
	 * @return the border insets for this component
	 */
	public Insets getBorderInsets()
	{
	   Border border = getBorder();
	   if (border != null)
		   return border.getBorderInsets( this );
	   else
		   return ZERO_INSET;
	}


	/*
	 * Overridden methods of JComponent
	 */
	
	public void paint( Graphics g )
	{
		super.paint( g );
		
		/* set a flag to say a 'paint' has been requested */
		_paintRequested = true;
		
		/* try to paint the image
		   (synchronize to ensure image is not changed between null-ptr-check & drawImage) */
		boolean hasBeenPainted;
		synchronized (this)
		{
			if (_displayImage == null)
			{
				hasBeenPainted = false;
				//System.out.println( "JImage::paint - no image to paint" );
			}
			else
			{
				hasBeenPainted = g.drawImage( _displayImage,
				                              _displayPosition.x,
				                              _displayPosition.y,
				                              new RepaintImageObserver( this, _displayImage ) );
				//System.out.println( "JImage::paint - image painted: " + hasBeenPainted );
			}
		}
		
		if (hasBeenPainted)
		{
			// TODO cancel repaint image observer
			/* image has been successfully rendered, clear the paint request flag */
			_paintRequested = false;
		}
	}
	
	
	/*
	 * Static methods
	 */
	
	/**
	 * Checks if a file is an image of a known format.
	 * <br />
	 * This is done by looking at the file's extension.
	 * <br />
	 * <br />
	 * @param file the file to be tested
	 * @return <code>true</code> if the file has an known & supported extension
	 */
	public static boolean isSupportedImage( File file )
	{
		return isSupportedImage( file.getName() );
	}
	
	/**
	 * Checks if a file is an image of a known format.
	 * <br />
	 * This is done by looking at the file's extension.
	 * <br />
	 * <br />
	 * @param fileName the name of the file to be tested
	 * @return <code>true</code> if the file has an known & supported extension
	 */
	public static boolean isSupportedImage( String fileName )
	{
		/* look for a file extension */
		int fileExtnIndex = fileName.lastIndexOf( '.' );
		if (fileExtnIndex == -1)
			return false;
		
		String fileExtension = fileName.substring( fileExtnIndex );
		
		/* look for extension in list of supported types */
		for (int i=0; i<SUPPORTED_IMAGE_EXTNS.length; i++)
		{
			if (fileExtension.equalsIgnoreCase( SUPPORTED_IMAGE_EXTNS[i] ))
				return true;
		}
		
		return false;
	}

}


/*
 * Change History
 *
 * $Log: JImage.java,v $
 * Revision 1.5  2006/09/06 21:07:54  Steve
 * Added legal disclaimer.
 *
 * Revision 1.4  2006/09/02 17:55:04  Steve
 * Extracted dimension image observer out to a separate class.
 *
 * Revision 1.3  2006/09/02 15:13:03  Steve
 * Added braces around if branches.
 * Removed debug messages.
 *
 * Revision 1.2  2006/07/10 21:13:20  Steve
 * Initial revision.
 *
 */
