/*
 * JImageTest.java
 * Part of the B'am library - a collection of general-purpose utility classes.
 * 
 * Copyright (C)2006 Stephen Battey
 * Contact: resource@lucidviews.net
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

package net.lucidviews.util.gui.swing;

import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.UIManager;

import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.event.WindowEvent;
import java.awt.AWTEvent;
import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Image;
import java.awt.Toolkit;


/**
 * Test suite for the <code>JImage</code> class.
 * <br />
 * <br />
 * @author Stephen Battey
 * @since 1.0
 */
public class JImageTest extends JFrame
{
	
	/*
	 * Constants
	 */
	
	/** The path of the image used in the test. */
	protected static final String TEST_IMAGE_PATH = "Test/JImage/KingsNorton.jpg";
	
	/** The image object created from the file at the TEST_IMAGE_PATH. */
	protected static final Image TEST_IMAGE = Toolkit.getDefaultToolkit().createImage( TEST_IMAGE_PATH );
	
	/** A list of scale settings that will be tested. */
	protected static final Dimension[] SCALE_SETTINGS = new Dimension[]{ JImage.MAINTAIN_SIZE,
	                                                                     JImage.FIT_TO_AREA,
	                                                                     JImage.FIT_TO_AREA_MAINTAIN_ASPECT,
	                                                                     JImage.SHRINK_TO_AREA_MAINTAIN_ASPECT };
	
	/** A list of descriptions for each of the scale settings to be tested. */
	protected static final String[] SCALE_SETTINGS_DESC = new String[]{ "maintain size",
	                                                                    "fit to area",
	                                                                    "fit to area, maintain aspect",
	                                                                    "shrink to area, maintain aspect" };
	
	
	/*
	 * Instance variables
	 */
	
	/** The component that we're testing. */
	private JImage _imagePane;
	
	/** Button to control the test. */
	private JButton _changeButton;
	
	/** Button to force the image component to refresh. */
	private JButton _repaintButton;
	
	/** The index of the scale setting currently being used. */
	private int _scaleSettingIndex;
	
	
	/*
	 * Constructor(s)
	 */
	
	/**
	 * The default constructor.
	 */
	public JImageTest()
	{
		frameInit();
		
		// enable events so we can detect when the window is closed
		enableEvents( AWTEvent.WINDOW_EVENT_MASK );
		
		_imagePane = new JImage();
		_imagePane.setPreferredSize( new Dimension( 300, 400 ) );
		
		_changeButton = new JButton();
		_changeButton.addActionListener( new ChangeButtonHandler( this ) );
		
		_repaintButton = new JButton( "Repaint" );
		_repaintButton.addActionListener( new RepaintButtonHandler( this ) );
		
		// initialise the image (and button text)
		_scaleSettingIndex = -1;
		nextImage();
		
		// place the components in this frame (window)
		Container contentPane = this.getContentPane();
		contentPane.setLayout( new BorderLayout() );
		contentPane.add( _imagePane, BorderLayout.CENTER );
		contentPane.add( _changeButton, BorderLayout.NORTH );
		contentPane.add( _repaintButton, BorderLayout.SOUTH );
	}
	
	
	/*
	 * Methods
	 */
	
	/**
	 * Displays the next image, changing the scale setting to the new test
	 * value.
	 * <br />
	 * The JImage pane and the toggle button must be created before calling this
	 * method.
	 */
	public void nextImage()
	{
		_scaleSettingIndex ++;
		if (_scaleSettingIndex >= SCALE_SETTINGS.length)
			_scaleSettingIndex = 0;
		
		_imagePane.setImage( TEST_IMAGE_PATH, SCALE_SETTINGS[_scaleSettingIndex], Image.SCALE_FAST );
		_changeButton.setText( "(" + SCALE_SETTINGS_DESC[_scaleSettingIndex] + ") ... NEXT" );
	}

	/**
	 * Process an incoming window events.
	 * <br />
	 * Overridden so we can quit the application process when window is closed.
	 * <br />
	 * <br />
	 * @param winEvent windows event object
	 */
	protected void processWindowEvent( WindowEvent winEvent )
	{
		super.processWindowEvent( winEvent );

		if (winEvent.getID() == WindowEvent.WINDOW_CLOSING)
		/* if window is being closed, exit the program */
		{
			System.exit( 0 );
		}
	}
	
	
	/*
	 * Event handler class
	 */
	
	/**
	 * An event handler assigned to the 'next' button that will trigger the
	 * <code>nextImage</code> method of the test frame.
	 */
	public static class ChangeButtonHandler implements ActionListener
	{
		public JImageTest _testFrame;
		
		public ChangeButtonHandler( JImageTest testFrame )
		{
			_testFrame = testFrame;
		}

	    public void actionPerformed(ActionEvent e)
	    {
	    	_testFrame.nextImage();
	    }
	}
	
	/**
	 * An event handler assigned to the 'next' button that will trigger the
	 * <code>nextImage</code> method of the test frame.
	 */
	public static class RepaintButtonHandler implements ActionListener
	{
		public JImageTest _testFrame;
		
		public RepaintButtonHandler( JImageTest testFrame )
		{
			_testFrame = testFrame;
		}

	    public void actionPerformed(ActionEvent e)
	    {
	    	_testFrame._imagePane.repaint();
	    }
	}
	
	
	/*
	 * Static methods
	 */
	
	/**
	 * Main entry point for starting the game in an application window.
	 * <p>
	 * @param args application arguments
	 */
	public static void main( String[] args )
	{
		try
		{
			/* set look and feel */
			UIManager.setLookAndFeel( UIManager.getSystemLookAndFeelClassName() );
			
			/* create the application window */
			JImageTest appnWindow = new JImageTest();
			appnWindow.pack();
			appnWindow.setVisible( true );
		}
		catch (Throwable e)
		{
			/* critical error: log it and write to console */
			e.printStackTrace();
			
			/* explicitly exit process to ensure threads controlling window are terminated */
			System.exit( 0 );
		}
	}
	
}


/*
 * Change History
 *
 * $Log: JImageTest.java,v $
 * Revision 1.3  2006/09/06 21:07:54  Steve
 * Added legal disclaimer.
 *
 * Revision 1.2  2006/09/02 15:10:22  Steve
 * Added a button to trigger a repaint.
 * Loaded the test image into an Image object, which can be used in the test instead of the path of the image file.
 *
 * Revision 1.1  2006/07/29 16:35:01  Steve
 * Initial revision.
 *
 */
