/*
 * JLettersPanel.java
 * Part of the B'am library - a collection of general-purpose utility classes.
 * 
 * Copyright (C)2006 Stephen Battey
 * Contact: resource@lucidviews.net
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

package net.lucidviews.util.gui.swing;

import javax.swing.border.EmptyBorder;
import javax.swing.border.Border;
import javax.swing.JLabel;
import javax.swing.JPanel;
import java.awt.FlowLayout;
import java.awt.LayoutManager;
import java.awt.Color;
import java.awt.Font;


/**
 * A panel that displays a row of letters.
 * <br />
 * <br />
 * @author Stephen Battey
 * @version $Revision: 1.3 $
 * @since 1.0
 */
public class JLettersPanel extends JPanel
{
	
	/*
	 * Instance variables
	 */
	
	/** The border, if no other is specified, to use around each text component
	 *  that displays a letter. */
	protected Border   _defaultBorder = new EmptyBorder( 2, 2, 2, 2 );
	
	/** Text components, one for each letter. */
	protected JLabel[] _letters = null;
	
	
	/*
	 * Constructor(s)
	 */
	
	/**
	 * Default constructor.
	 * <br />
	 * <br />
	 */
	public JLettersPanel()
	{
		super( new FlowLayout( FlowLayout.CENTER, 0, 0 ) );
	}
	
	/**
	 * Default constructor.
	 * <br />
	 * <br />
	 * @param isDoubleBuffered a boolean, true for double-buffering, which uses
	 *                         additional memory space to achieve fast,
	 *                         flicker-free updates
	 */
	public JLettersPanel( boolean isDoubleBuffered )
	{
		super( new FlowLayout( FlowLayout.CENTER, 0, 0 ), isDoubleBuffered );
	}
	
	
	/*
	 * Overridden methods of Container base class
	 */
	
	public void setLayout( LayoutManager layoutManager )
	{
		if (!(layoutManager instanceof FlowLayout))
			/* this panel always uses a FlowLayout */
			throw new UnsupportedOperationException( "The layout manager for a JLettersPanel cannot be changed." );
		
		super.setLayout( layoutManager );
	}
	
	
	/*
	 * Methods
	 */
	
	/**
	 * Sets the alignment for the row of letters within the whole panel.
	 * Possible values are
	 * <ul>
	 * <li><code>FlowLayout.LEFT</code></li>
	 * <li><code>FlowLayout.RIGHT</code></li>
	 * <li><code>FlowLayout.CENTER</code></li>
	 * <li><code>FlowLayout.LEADING</code></li>
	 * <li><code>FlowLayout.TRAILING</code></li>
	 * </ul>
 	 * @param align one of the alignment values shown above
 	 */
	public void setAlignment( int align )
	{
		((FlowLayout)getLayout()).setAlignment( align );
	}
	
	/**
	 * Set the border used around each letter.
	 * <br />
	 * The border of all current letters will be updated.
	 * <br />
	 * <br />
	 * @param border the border around each letter
	 */
	public void setLetterBorder( Border border )
	{
		_defaultBorder = border;
		
		/* update border for letters currently on display */
		if (_letters != null)
		{
			for (int i=0; i<_letters.length; i++)
				_letters[i].setBorder( _defaultBorder );
		}
	}
	
	/**
	 * Set the letters to be displayed.
	 * <br />
	 * The current set of letters will be removed and replaced with those in the
	 * <code>String</code>.
	 * <br />
	 * <br />
	 * @param letters a string of letters
	 */
	public void setLetters( String letters )
	{
		/* remove all letters currently on display */
		removeAll();
		
		/* create the labels */
		_letters = new JLabel[letters.length()];
		for (int i=0; i<letters.length(); i++)
		{
			JLabel newLabel = new JLabel( letters.substring( i, i+1 ) );
			newLabel.setBorder( _defaultBorder );
			newLabel.setFont( null );
			newLabel.setForeground( null );
			newLabel.setBackground( null );
			
			_letters[i] = newLabel;
			add( newLabel );
		}
	}
	
	/**
	 * Set the letters to be displayed.
	 * <br />
	 * The current set of letters will be removed and replaced with a sequence
	 * of a single letter.
	 * <br />
	 * <br />
	 * @param letter the letter to be used
	 * @param count the number of letters to be added to the panel
	 */
	public void setLetters( char letter, int count )
	{
		/* build a string with the specified number of letters */
		StringBuffer buf = new StringBuffer( count );
		for (int i=0; i<count; i++)
			buf.append( letter );
		
		setLetters( buf.toString() );
	}

	/**
	 * Change the letter shown at a given position.
	 * <br />
	 * If the index goes beyond the list of letters currently displayed on this
	 * panel then an exception will be thrown.
	 * <br />
	 * <br />
	 * @param letter the letter to be displayed
	 * @param index the index within the panel where the letter is to appear
	 * @throws IndexOutOfBoundsException there is no letter at the specified
	 *                                   index
	 */
	public void setLetter( char letter, int index ) throws IndexOutOfBoundsException
	{
		getLetter( index ).setText( Character.toString( letter ) );
	}
	
	/**
	 * Change the letter shown at a given position.
	 * <br />
	 * If the index goes beyond the list of letters currently displayed on this
	 * panel then an exception will be thrown.
	 * <br />
	 * <br />
	 * @param letter the letter to be displayed
	 * @param index the index within the panel where the letter is to appear
	 * @param font the font to be used to render the letter
	 * @param colour the colour of the letter when it is rendered
	 * @throws IndexOutOfBoundsException there is no letter at the specified
	 *                                   index
	 */
	public void setLetter( char letter, int index, Font font, Color colour ) throws IndexOutOfBoundsException
	{
		JLabel letterLabel = getLetter( index );

		/* update the letter */
		letterLabel.setText( Character.toString( letter ) );
		letterLabel.setFont( font );
		letterLabel.setForeground( colour );
	}
	
	
	/*
	 * Local methods
	 */
	
	/**
	 * Returns the label showing the letter at the given index.
	 * <br />
	 * If the letter does not exist an <code>IndexOutOfBoundsException</code>
	 * is thrown.
	 * <br />
	 * <br />
	 * @param index the index of the letter
	 * @return the <code>JLabel</code> that renders the letter
	 * @throws IndexOutOfBoundsException there is no letter at the specified
	 *                                   index
	 */
	protected JLabel getLetter( int index ) throws IndexOutOfBoundsException
	{
		/* verify the letter exists */
		if (   (_letters == null)
		    || (_letters.length <= index)  )
			throw new IndexOutOfBoundsException( "Letter at index " + index + " does not exist (letters are " + _letters + ")." );
		
		return _letters[index];
	}
	
}


/*
 * Change History
 *
 * $Log: JLettersPanel.java,v $
 * Revision 1.3  2006/09/06 21:07:54  Steve
 * Added legal disclaimer.
 *
 * Revision 1.2  2006/06/08 19:53:49  Steve
 * Brought into Bam library:
 *  - moved into lucidviews package
 *  - added CVS headers/footers
 *
 */
