/*
 * WindowUtils.java
 * Part of the B'am library - a collection of general-purpose utility classes.
 * 
 * Copyright (C)2006 Stephen Battey
 * Contact: resource@lucidviews.net
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

package net.lucidviews.util.gui.swing;

import net.lucidviews.util.Properties;

import javax.swing.JFrame;
import java.awt.Dimension;
import java.awt.MediaTracker;
import java.awt.Image;
import java.awt.Toolkit;
import java.awt.event.WindowListener;
import java.awt.event.WindowEvent;
import java.io.IOException;
import java.io.FileNotFoundException;


/**
 * Utilities for setting window properties.
 * <br />
 * <br />
 * @author Stephen Battey
 * @version $Revision: 1.2 $
 * @since 1.0
*/
public class WindowUtils implements WindowListener
{
	
	/*
	 * Local constants
	 */
	
	/** Suffix for window width property name. */
	private static final String WINDOW_WIDTH_PROPERTY_NAME = ".window.width";
	
	/** Suffix for window height property name. */
	private static final String WINDOW_HEIGHT_PROPERTY_NAME = ".window.height";
	
	/** Suffix for window x-position property name. */
	private static final String WINDOW_XPOS_PROPERTY_NAME = ".window.xpos";
	
	/** Suffix for window y-position property name. */
	private static final String WINDOW_YPOS_PROPERTY_NAME = ".window.ypos";
	
	
	/*
	 * Instance variables
	 */
	
	/** A reference to the window that the properties are applicable to. */
	private JFrame _targetWindow;
	
	/** The properties for the window. */
	private Properties _appnProperties;
	
	/** The prefix for all property names (should include the name of the
	 *  application). */
	private String _appnPropertyPrefix;
	
	
	/*
	 * Constructor(s)
	 */
	
	/**
	 * Create a utility suite for a given application window.
	 * <br />
	 * <br />
	 * @param target the window to be controlled
	 */
	public WindowUtils( JFrame target )
	{
		_targetWindow = target;
		_appnProperties = null;
		_appnPropertyPrefix = null;
	}
	
	/**
	 * Create a utility suite for a given application window.
	 * <br />
	 * <br />
	 * @param target the window to be controlled
	 * @param properties data model containing the window's properties
	 * @param propertyPrefix the prefix to the names of all the window's properties
	 */
	public WindowUtils( JFrame target, Properties properties, String propertyPrefix )
	{
		_targetWindow = target;
		_appnProperties = properties;
		_appnPropertyPrefix = propertyPrefix;
	}
	
	/**
	 * Create a utility suite for a given application window.
	 * <br />
	 * <br />
	 * @param target the window to be controlled
	 * @param propertiesFileRef URL of a file that stores the window's properties
	 * @param propertyPrefix the prefix to the names of all the window's properties
	 * @throws FileNotFoundException the file location specified by propertiesFileRef does not exist
	 * @throws IOException the file at the specified location could not be read
	 */
	public WindowUtils( JFrame target, String propertiesFileRef, String propertyPrefix ) throws FileNotFoundException, IOException
	{
		_targetWindow = target;
		_appnProperties = new Properties( propertiesFileRef );
		_appnPropertyPrefix = propertyPrefix;
	}
	
	
	/*
	 * Local methods
	 */
	
	private void storeWindowSize0() throws IOException
	{
		_appnProperties.setLongProperty( _appnPropertyPrefix + WINDOW_WIDTH_PROPERTY_NAME, _targetWindow.getWidth() );
		_appnProperties.setLongProperty( _appnPropertyPrefix + WINDOW_HEIGHT_PROPERTY_NAME, _targetWindow.getHeight() );
	}
	
	private void storeWindowLocation0() throws IOException
	{
		_appnProperties.setLongProperty( _appnPropertyPrefix + WINDOW_XPOS_PROPERTY_NAME, _targetWindow.getX() );
		_appnProperties.setLongProperty( _appnPropertyPrefix + WINDOW_YPOS_PROPERTY_NAME, _targetWindow.getY() );
	}
	
	
	/*
	 * Methods
	 */
	
	/**
	 * Set the icon for this window.
	 * <br />
	 * The icon will be displayed in the top left corner of the window.
	 * <br />
	 * This method will automatically load the image from the URL, waiting for the
	 * image data to load before returning. This ensure the icon is ready for display
	 * before you go on to make the window visible by calling <code>setVisible</code>.
	 * <br />
	 * <br />
	 * @param iconImageURL the URL to the image file
	 */
	public void setIcon( String iconImageURL )
	{
		/* load the image */
		Image windowIcon = Toolkit.getDefaultToolkit().createImage( iconImageURL );
		MediaTracker tracker = new MediaTracker( _targetWindow );
		tracker.addImage( windowIcon, 0 );
		try
		{
			tracker.waitForAll();
		}
		catch (InterruptedException e)
		{
			// respect the interruption
			// - abandon wait and assume the image is not required
		}
		
		_targetWindow.setIconImage( windowIcon );
	}
	
	/**
	 * Set the size of the window.
	 * <br />
	 * The size will be first taken from the application properties table. If there is
	 * no properties table or the properties do not contain a width and height, the
	 * window will take on the preferred size of the internal components of the window.
	 * <br />
	 * <br />
	 */
	public void setWindowSize()
	{
		try
		{
			// set dimension of window from last save
			_targetWindow.setSize( _appnProperties.getLongProperty( _appnPropertyPrefix + WINDOW_WIDTH_PROPERTY_NAME ).intValue(),
			                       _appnProperties.getLongProperty( _appnPropertyPrefix + WINDOW_HEIGHT_PROPERTY_NAME ).intValue() );
		
			// verify the content of the window fits
			_targetWindow.validate();
		}
		catch (Throwable error)
		{
			// In the case of any error (e.g. _appnProperties is null or the
			// width or height are not stored) pack the contents into the main
			// frame if the window has no fixed size.
			_targetWindow.pack();
		}
	}
	
	/**
	 * Set the position of the window.
	 * <br />
	 * The (x and y) position will be first taken from the application properties table.
	 * If there is no properties table or the properties do not contain an x and y co-ordinate,
	 * the window will be placed in the centre of the screen.
	 * <br />
	 * <br />
	 */
	public void setWindowLocation()
	{
		int xPos;
		int yPos;
		
		try
		{
			// load position of window from last save
			xPos = _appnProperties.getLongProperty( _appnPropertyPrefix + WINDOW_XPOS_PROPERTY_NAME ).intValue();
			yPos = _appnProperties.getLongProperty( _appnPropertyPrefix + WINDOW_YPOS_PROPERTY_NAME ).intValue();
		}
		catch (Throwable error)
		{
			// In the case of any error (e.g. _appnProperties is null or the x
			// or y position are not stored) place the window in the centre of
			// the screen.
			Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
			Dimension frameSize = _targetWindow.getSize();
			
			if (frameSize.width > screenSize.width)
				xPos = 0;
			else
				xPos = (screenSize.width - frameSize.width) / 2;
			
			if (frameSize.height > screenSize.height)
				yPos = 0;
			else
				yPos = (screenSize.height - frameSize.height) / 2;
		}
		
		_targetWindow.setLocation( xPos, yPos );
	}
	
	/**
	 * Store the size of the window.
	 * <br />
	 * If there is no properties table this operation is ignored.
	 * Otherwise, the current width and height of the window are saved to the file
	 * where the properties were originally read from.
	 * <br />
	 * <br />
	 * @throws IOException error writing the size properties to the properties file
	 */
	public void storeWindowSize() throws IOException
	{
		if (_appnProperties != null)
		{
			storeWindowSize0();
			_appnProperties.store();
		}
	}
	
	/**
	 * Store the position of the window.
	 * <br />
	 * If there is no properties table this operation is ignored.
	 * Otherwise, the current x and y co-ordinates of the window are saved to the file
	 * where the properties were originally read from.
	 * <br />
	 * <br />
	 * @throws IOException error writing the location properties to the properties file
	 */
	public void storeWindowLocation() throws IOException
	{
		if (_appnProperties != null)
		{
			storeWindowLocation0();
			_appnProperties.store();
		}
	}
	
	/**
	 * Store the size and position of the window.
	 * <br />
	 * If there is no properties table this operation is ignored.
	 * Otherwise, the current width, height, x co-ordinate and y co-ordinate of the
	 * window are saved to the file where the properties were originally read from.
	 * <br />
	 * <br />
	 * @throws IOException error writing the window properties to the properties file
	 */
	public void storeWindowProperties() throws IOException
	{
		if (_appnProperties != null)
		{
			storeWindowSize0();
			storeWindowLocation0();
			_appnProperties.store();
		}
	}
	
	/**
	 * Automatically store the properties of this window when the window is closed.
	 * <br />
	 * The utils object will listen for the close-window event and save the window
	 * properties, ready for the next time the application is run.
	 * <br />
	 * <br />
	 */
	public void autoStoreWindowProperties()
	{
		_targetWindow.addWindowListener( this );
	}
	
	
	/*
	 * Implementation of WindowListener interface
	 */
	
	public void windowOpened( WindowEvent e )
	{
	}
	
	public void windowClosing( WindowEvent e )
	{
		try
		{
			storeWindowProperties();
		}
		catch (IOException ex)
		{
			// Not much we could do with this error.
			// The information could not be saved, but as the window is about to
			// close there is nothing we can do to inform the user.
			System.err.println( "Unable to save window properties. " + ex.getMessage() );
		}
	}
	
	public void windowClosed( WindowEvent e )
	{
	}
	
	public void windowIconified( WindowEvent e )
	{
	}
	
	public void windowDeiconified( WindowEvent e )
	{
	}
	
	public void windowActivated( WindowEvent e )
	{
	}
	
	public void windowDeactivated( WindowEvent e )
	{
	}
	
}


/*
 * Change History
 *
 * $Log: WindowUtils.java,v $
 * Revision 1.2  2006/09/06 21:07:54  Steve
 * Added legal disclaimer.
 *
 * Revision 1.1  2006/08/07 19:45:26  Steve
 * Initial revision.
 *
 */
